# ***************************************************************************
# *   Copyright (c) 2019 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

# to run the example use:
"""
from femexamples.material_nl_platewithhole import setup
setup()

"""

# Nonlinear material example, plate with hole.

# https://forum.freecadweb.org/viewtopic.php?f=24&t=31997&start=30
# https://forum.freecadweb.org/viewtopic.php?t=33974&start=90
# https://forum.freecadweb.org/viewtopic.php?t=35893

# plate: 400x200x10 mm
# hole: diameter 100 mm (half cross section)
# load: 130 MPa tension
# linear material: Steel, E = 210000 MPa, my = 0.3
# nonlinear material: '240.0, 0.0' to '270.0, 0.025'
# TODO nonlinear material: give more information, use values from harry
# TODO compare results with example from HarryvL

import FreeCAD
from FreeCAD import Vector as vec

import Fem
import ObjectsFem
import Part
from Part import makeCircle as ci
from Part import makeLine as ln

mesh_name = "Mesh"  # needs to be Mesh to work with unit tests


def init_doc(doc=None):
    if doc is None:
        doc = FreeCAD.newDocument()
    return doc


def setup(doc=None, solvertype="ccxtools"):
    # setup model

    if doc is None:
        doc = init_doc()

    # geometry objects
    v1 = vec(-200, -100, 0)
    v2 = vec(200, -100, 0)
    v3 = vec(200, 100, 0)
    v4 = vec(-200, 100, 0)
    l1 = ln(v1, v2)
    l2 = ln(v2, v3)
    l3 = ln(v3, v4)
    l4 = ln(v4, v1)
    v5 = vec(0, 0, 0)
    c1 = ci(50, v5)
    face = Part.makeFace([Part.Wire([l1, l2, l3, l4]), c1], "Part::FaceMakerBullseye")
    geom_obj = doc.addObject("Part::Feature", "Hole_Plate")
    geom_obj.Shape = face.extrude(vec(0, 0, 10))
    doc.recompute()

    if FreeCAD.GuiUp:
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")

    # solver
    if solvertype == "calculix":
        solver = analysis.addObject(
            ObjectsFem.makeSolverCalculix(doc, "SolverCalculiX")
        )[0]
    elif solvertype == "ccxtools":
        solver = analysis.addObject(
            ObjectsFem.makeSolverCalculixCcxTools(doc, "CalculiXccxTools")
        )[0]
        solver.WorkingDir = u""
    if solvertype == "calculix" or solvertype == "ccxtools":
        solver.SplitInputWriter = False
        solver.AnalysisType = "static"
        solver.GeometricalNonlinearity = "linear"
        solver.ThermoMechSteadyState = False
        solver.MatrixSolverType = "default"
        solver.IterationsControlParameterTimeUse = False
        solver.GeometricalNonlinearity = 'nonlinear'
        solver.MaterialNonlinearity = 'nonlinear'

    # linear material
    matprop = {}
    matprop["Name"] = "CalculiX-Steel"
    matprop["YoungsModulus"] = "210000 MPa"
    matprop["PoissonRatio"] = "0.30"
    matprop["Density"] = "7900 kg/m^3"
    material = analysis.addObject(
        ObjectsFem.makeMaterialSolid(doc, "Material_lin")
    )[0]
    material.Material = matprop

    # nonlinear material
    nonlinear_material = analysis.addObject(
        ObjectsFem.makeMaterialMechanicalNonlinear(doc, material, "Material_nonlin")
    )[0]
    nonlinear_material.YieldPoint1 = '240.0, 0.0'
    nonlinear_material.YieldPoint2 = '270.0, 0.025'
    # check solver attributes, Nonlinearity needs to be set to nonlinear

    # fixed_constraint
    fixed_constraint = analysis.addObject(
        ObjectsFem.makeConstraintFixed(doc, "ConstraintFixed")
    )[0]
    fixed_constraint.References = [(geom_obj, "Face4")]

    # force constraint
    pressure_constraint = doc.Analysis.addObject(
        ObjectsFem.makeConstraintPressure(doc, "ConstraintPressure")
    )[0]
    pressure_constraint.References = [(geom_obj, "Face2")]
    pressure_constraint.Pressure = 130.0
    pressure_constraint.Reversed = True

    # mesh
    from .meshes.mesh_platewithhole_tetra10 import create_nodes, create_elements
    fem_mesh = Fem.FemMesh()
    control = create_nodes(fem_mesh)
    if not control:
        FreeCAD.Console.PrintError("Error on creating nodes.\n")
    control = create_elements(fem_mesh)
    if not control:
        FreeCAD.Console.PrintError("Error on creating elements.\n")
    femmesh_obj = analysis.addObject(
        doc.addObject("Fem::FemMeshObject", mesh_name)
    )[0]
    femmesh_obj.FemMesh = fem_mesh

    doc.recompute()
    return doc



# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""This module provides the object code for Draft Label.
"""
## @package label
# \ingroup DRAFT
# \brief This module provides the object code for Draft Label.

import FreeCAD as App
import math
from PySide.QtCore import QT_TRANSLATE_NOOP
import DraftGeomUtils
import draftutils.gui_utils as gui_utils
import draftutils.utils as utils
from draftobjects.draft_annotation import DraftAnnotation

if App.GuiUp:
    from draftviewproviders.view_label import ViewProviderLabel



def make_label(targetpoint=None, target=None, direction=None,
               distance=None, labeltype=None, placement=None):
    """
    make_label(targetpoint, target, direction, distance, labeltype, placement)
    
    Function to create a Draft Label annotation object

    Parameters
    ----------
    targetpoint : App::Vector
            To be completed

    target  : LinkSub
            To be completed

    direction : String
            Straight direction of the label
            ["Horizontal","Vertical","Custom"]

    distance : Quantity
            Length of the straight segment of label leader line 

    labeltype : String
            Label type in
            ["Custom","Name","Label","Position",
            "Length","Area","Volume","Tag","Material"]

    placement : Base::Placement
            To be completed

    Returns
    -------
    obj :   App::DocumentObject
            Newly created label object
    """
    obj = App.ActiveDocument.addObject("App::FeaturePython", 
                                       "dLabel")
    Label(obj)
    if App.GuiUp:
        ViewProviderLabel(obj.ViewObject)
    if targetpoint:
        obj.TargetPoint = targetpoint
    if target:
        obj.Target = target
    if direction:
        obj.StraightDirection = direction
    if distance:
        obj.StraightDistance = distance
    if labeltype:
        obj.LabelType = labeltype
    if placement:
        obj.Placement = placement

    if App.GuiUp:
        gui_utils.format_object(obj)
        gui_utils.select(obj)

    return obj



class Label(DraftAnnotation):
    """The Draft Label object"""

    def __init__(self, obj):

        super(Label, self).__init__(obj, "Label")

        self.init_properties(obj)

        obj.Proxy = self


    def init_properties(self, obj):
        """Add properties to the object and set them"""

        obj.addProperty("App::PropertyPlacement",
                        "Placement",
                        "Base",
                        QT_TRANSLATE_NOOP("App::Property",
                                          "The placement of this object"))

        obj.addProperty("App::PropertyDistance",
                        "StraightDistance",
                        "Base",
                        QT_TRANSLATE_NOOP("App::Property",
                                          "The length of the straight segment"))

        obj.addProperty("App::PropertyVector",
                        "TargetPoint",
                        "Base",
                        QT_TRANSLATE_NOOP("App::Property",
                                          "The point indicated by this label"))

        obj.addProperty("App::PropertyVectorList",
                        "Points",
                        "Base",
                        QT_TRANSLATE_NOOP("App::Property",
                                          "The points defining the label polyline"))

        obj.addProperty("App::PropertyEnumeration",
                        "StraightDirection",
                        "Base",
                        QT_TRANSLATE_NOOP("App::Property",
                                          "The direction of the straight segment"))

        obj.addProperty("App::PropertyEnumeration",
                        "LabelType",
                        "Base",
                        QT_TRANSLATE_NOOP("App::Property",
                                          "The type of information shown by this label"))

        obj.addProperty("App::PropertyLinkSub",
                        "Target",
                        "Base",
                        QT_TRANSLATE_NOOP("App::Property",
                                          "The target object of this label"))

        obj.addProperty("App::PropertyStringList",
                        "CustomText",
                        "Base",
                        QT_TRANSLATE_NOOP("App::Property",
                                          "The text to display when type is set to custom"))

        obj.addProperty("App::PropertyStringList",
                        "Text",
                        "Base",
                        QT_TRANSLATE_NOOP("App::Property",
                                          "The text displayed by this label"))

        obj.StraightDirection = ["Horizontal","Vertical","Custom"]
        obj.LabelType = ["Custom","Name","Label","Position",
                         "Length","Area","Volume","Tag","Material"]
        obj.setEditorMode("Text",1)
        obj.StraightDistance = 1
        obj.TargetPoint = App.Vector(2,-1,0)
        obj.CustomText = "Label"


    def execute(self,obj):
        '''Do something when recompute object'''

        if obj.StraightDirection != "Custom":
            p1 = obj.Placement.Base
            if obj.StraightDirection == "Horizontal":
                p2 = App.Vector(obj.StraightDistance.Value,0,0)
            else:
                p2 = App.Vector(0,obj.StraightDistance.Value,0)
            p2 = obj.Placement.multVec(p2)
            # p3 = obj.Placement.multVec(obj.TargetPoint)
            p3 = obj.TargetPoint
            obj.Points = [p1,p2,p3]
        if obj.LabelType == "Custom":
            if obj.CustomText:
                obj.Text = obj.CustomText
        elif obj.Target and obj.Target[0]:
            if obj.LabelType == "Name":
                obj.Text = [obj.Target[0].Name]
            elif obj.LabelType == "Label":
                obj.Text = [obj.Target[0].Label]
            elif obj.LabelType == "Tag":
                if hasattr(obj.Target[0],"Tag"):
                    obj.Text = [obj.Target[0].Tag]
            elif obj.LabelType == "Material":
                if hasattr(obj.Target[0],"Material"):
                    if hasattr(obj.Target[0].Material,"Label"):
                        obj.Text = [obj.Target[0].Material.Label]
            elif obj.LabelType == "Position":
                p = obj.Target[0].Placement.Base
                if obj.Target[1]:
                    if "Vertex" in obj.Target[1][0]:
                        p = obj.Target[0].Shape.Vertexes[int(obj.Target[1][0][6:])-1].Point
                obj.Text = [App.Units.Quantity(x,App.Units.Length).UserString for x in tuple(p)]
            elif obj.LabelType == "Length":
                if hasattr(obj.Target[0],'Shape'):
                    if hasattr(obj.Target[0].Shape,"Length"):
                        obj.Text = [App.Units.Quantity(obj.Target[0].Shape.Length,App.Units.Length).UserString]
                    if obj.Target[1] and ("Edge" in obj.Target[1][0]):
                        obj.Text = [App.Units.Quantity(obj.Target[0].Shape.Edges[int(obj.Target[1][0][4:])-1].Length,App.Units.Length).UserString]
            elif obj.LabelType == "Area":
                if hasattr(obj.Target[0],'Shape'):
                    if hasattr(obj.Target[0].Shape,"Area"):
                        obj.Text = [App.Units.Quantity(obj.Target[0].Shape.Area,App.Units.Area).UserString.replace("^2","²")]
                    if obj.Target[1] and ("Face" in obj.Target[1][0]):
                        obj.Text = [App.Units.Quantity(obj.Target[0].Shape.Faces[int(obj.Target[1][0][4:])-1].Area,App.Units.Area).UserString]
            elif obj.LabelType == "Volume":
                if hasattr(obj.Target[0],'Shape'):
                    if hasattr(obj.Target[0].Shape,"Volume"):
                        obj.Text = [App.Units.Quantity(obj.Target[0].Shape.Volume,App.Units.Volume).UserString.replace("^3","³")]


    def onChanged(self,obj,prop):
        '''Do something when a property has changed'''

        return


# ***************************************************************************
# *   (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>                  *
# *   (c) 2009, 2010 Ken Cline <cline@frii.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides tools for joining lines with the Draft Workbench.

It occasionally fails to join lines even if the lines
visually share a point. This is due to the underlying `joinWires` method
not handling the points correctly.

This is a rounding error in the comparison of the shared point;
a small difference will result in the points being considered different
and thus the lines not joining.

Test properly using `DraftVecUtils.equals` because then it will consider
the precision set in the Draft preferences.
"""
## @package gui_join
# \ingroup DRAFT
# \brief Provides tools for joining lines with the Draft Workbench.

from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCADGui as Gui
import Draft_rc
import draftguitools.gui_base_original as gui_base_original
import draftguitools.gui_tool_utils as gui_tool_utils
from draftutils.messages import _msg
from draftutils.translate import translate, _tr

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class Join(gui_base_original.Modifier):
    """Gui Command for the Join tool."""

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _tip = ("Joins the selected lines or polylines "
                "into a single object.\n"
                "The lines must share a common point at the start "
                "or at the end for the operation to succeed.")

        return {'Pixmap': 'Draft_Join',
                'Accel': "J, O",
                'MenuText': QT_TRANSLATE_NOOP("Draft_Join", "Join"),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Join", _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(Join, self).Activated(name=_tr("Join"))
        if not self.ui:
            return
        if not Gui.Selection.getSelection():
            self.ui.selectUi()
            _msg(translate("draft", "Select an object to join"))
            self.call = self.view.addEventCallback("SoEvent",
                                                   gui_tool_utils.selectObject)
        else:
            self.proceed()

    def proceed(self):
        """Proceed with execution of the command after proper selection.

        BUG: It occasionally fails to join lines even if the lines
        visually share a point. This is due to the underlying `joinWires`
        method not handling the points correctly.
        """
        if self.call:
            self.view.removeEventCallback("SoEvent", self.call)
        if Gui.Selection.getSelection():
            self.print_selection()
            Gui.addModule("Draft")
            _cmd = "Draft.joinWires"
            _cmd += "("
            _cmd += "FreeCADGui.Selection.getSelection()"
            _cmd += ")"
            _cmd_list = ['j = ' + _cmd,
                         'FreeCAD.ActiveDocument.recompute()']
            self.commit(translate("draft", "Join lines"),
                        _cmd_list)
        self.finish()

    def print_selection(self):
        """Print the selected items."""
        labels = []
        for obj in Gui.Selection.getSelection():
            labels.append(obj.Label)

        labels = ", ".join(labels)
        _msg(_tr("Selection:") + " {}".format(labels))


Gui.addCommand('Draft_Join', Join())



# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""This module provides the code for Draft make_wire function.
"""
## @package make_wire
# \ingroup DRAFT
# \brief This module provides the code for Draft make_wire function.

import FreeCAD as App

import DraftGeomUtils

from draftutils.gui_utils import format_object
from draftutils.gui_utils import select

from draftutils.utils import type_check

from draftobjects.wire import Wire
if App.GuiUp:
    from draftviewproviders.view_wire import ViewProviderWire


def make_wire(pointslist, closed=False, placement=None, face=None, support=None, bs2wire=False):
    """makeWire(pointslist,[closed],[placement])
    
    Creates a Wire object from the given list of vectors.  If face is
    true (and wire is closed), the wire will appear filled. Instead of
    a pointslist, you can also pass a Part Wire.

    Parameters
    ----------
    pointlist : [Base.Vector]
        List of points to create the polyline

    closed : bool
        If closed is True or first and last points are identical, 
        the created polyline will be closed.

    placement : Base.Placement
        If a placement is given, it is used.
    
    face : Bool
        If face is False, the rectangle is shown as a wireframe, 
        otherwise as a face.   

    support : 
        TODO: Describe
    
    bs2wire : bool
        TODO: Describe
    """
    if not App.ActiveDocument:
        App.Console.PrintError("No active document. Aborting\n")
        return

    import Part

    if not isinstance(pointslist, list):
        e = pointslist.Wires[0].Edges
        pointslist = Part.Wire(Part.__sortEdges__(e))
        nlist = []
        for v in pointslist.Vertexes:
            nlist.append(v.Point)
        if DraftGeomUtils.isReallyClosed(pointslist):
            closed = True
        pointslist = nlist

    if len(pointslist) == 0:
        print("Invalid input points: ", pointslist)
    #print(pointslist)
    #print(closed)
    
    if placement:
        type_check([(placement, App.Placement)], "make_wire")
        ipl = placement.inverse()
        if not bs2wire:
            pointslist = [ipl.multVec(p) for p in pointslist]

    if len(pointslist) == 2: 
        fname = "Line"
    else: 
        fname = "Wire"

    obj = App.ActiveDocument.addObject("Part::Part2DObjectPython", fname)
    Wire(obj)
    obj.Points = pointslist
    obj.Closed = closed
    obj.Support = support
    
    if face != None:
        obj.MakeFace = face

    if placement: 
        obj.Placement = placement

    if App.GuiUp:
        ViewProviderWire(obj.ViewObject)
        format_object(obj)
        select(obj)

    return obj


makeWire = make_wire


# -*- coding: utf-8 -*-

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2018 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import PathScripts.PathLog as PathLog
import PathScripts.PathSetupSheetOpPrototype as PathSetupSheetOpPrototype

from PySide import QtCore, QtGui

__title__ = "Setup Sheet Editor"
__author__ = "sliptonic (Brad Collette)"
__url__ = "http://www.freecadweb.org"
__doc__ = "Task panel editor for a SetupSheet"

# Qt translation handling
def translate(context, text, disambig=None):
    return QtCore.QCoreApplication.translate(context, text, disambig)

LOGLEVEL = False

if LOGLEVEL:
    PathLog.setLevel(PathLog.Level.DEBUG, PathLog.thisModule())
    PathLog.trackModule(PathLog.thisModule())
else:
    PathLog.setLevel(PathLog.Level.INFO, PathLog.thisModule())

class _PropertyEditor(object):
    '''Base class of all property editors - just outlines the TableView delegate interface.'''
    def __init__(self, prop):
        self.prop = prop

    def widget(self, parent):
        '''widget(parent) ... called by the delegate to get a new editor widget.
        Must be implemented by subclasses and return the widget.'''
        pass # pylint: disable=unnecessary-pass
    def setEditorData(self, widget):
        '''setEditorData(widget) ... called by the delegate to initialize the editor.
        The widget is the object returned by widget().
        Must be implemented by subclasses.'''
        pass # pylint: disable=unnecessary-pass
    def setModelData(self, widget):
        '''setModelData(widget) ... called by the delegate to store new values.
        Must be implemented by subclasses.'''
        pass # pylint: disable=unnecessary-pass

class _PropertyEnumEditor(_PropertyEditor):
    '''Editor for enumeration values - uses a combo box.'''

    def widget(self, parent):
        PathLog.track(self.prop.name, self.prop.getEnumValues())
        return QtGui.QComboBox(parent)

    def setEditorData(self, widget):
        widget.clear()
        widget.addItems(self.prop.getEnumValues())
        if self.prop.getValue():
            index = widget.findText(self.prop.getValue(), QtCore.Qt.MatchFixedString)
            if index >= 0:
                widget.setCurrentIndex(index)

    def setModelData(self, widget):
        self.prop.setValue(widget.currentText())


class _PropertyBoolEditor(_PropertyEditor):
    '''Editor for boolean values - uses a combo box.'''

    def widget(self, parent):
        return QtGui.QComboBox(parent)

    def setEditorData(self, widget):
        widget.clear()
        widget.addItems(['false', 'true'])
        if not self.prop.getValue() is None:
            index = 1 if self.prop.getValue() else 0
            widget.setCurrentIndex(index)

    def setModelData(self, widget):
        self.prop.setValue(widget.currentText() == 'true')

class _PropertyStringEditor(_PropertyEditor):
    '''Editor for string values - uses a line edit.'''

    def widget(self, parent):
        return QtGui.QLineEdit(parent)

    def setEditorData(self, widget):
        text = '' if self.prop.getValue() is None else self.prop.getValue()
        widget.setText(text)

    def setModelData(self, widget):
        self.prop.setValue(widget.text())

class _PropertyAngleEditor(_PropertyEditor):
    '''Editor for angle values - uses a line edit'''

    def widget(self, parent):
        return QtGui.QLineEdit(parent)

    def setEditorData(self, widget):
        quantity = self.prop.getValue()
        if quantity is None:
            quantity = FreeCAD.Units.Quantity(0, FreeCAD.Units.Angle)
        widget.setText(quantity.getUserPreferred()[0])

    def setModelData(self, widget):
        self.prop.setValue(FreeCAD.Units.Quantity(widget.text()))

class _PropertyLengthEditor(_PropertyEditor):
    '''Editor for length values - uses a line edit.'''

    def widget(self, parent):
        return QtGui.QLineEdit(parent)

    def setEditorData(self, widget):
        quantity = self.prop.getValue()
        if quantity is None:
            quantity = FreeCAD.Units.Quantity(0, FreeCAD.Units.Length)
        widget.setText(quantity.getUserPreferred()[0])

    def setModelData(self, widget):
        self.prop.setValue(FreeCAD.Units.Quantity(widget.text()))

class _PropertyPercentEditor(_PropertyEditor):
    '''Editor for percent values - uses a spin box.'''

    def widget(self, parent):
        return QtGui.QSpinBox(parent)

    def setEditorData(self, widget):
        widget.setRange(0, 100)
        value = self.prop.getValue()
        if value is None:
            value = 0
        widget.setValue(value)

    def setModelData(self, widget):
        self.prop.setValue(widget.value())

class _PropertyIntegerEditor(_PropertyEditor):
    '''Editor for integer values - uses a spin box.'''

    def widget(self, parent):
        return QtGui.QSpinBox(parent)

    def setEditorData(self, widget):
        value = self.prop.getValue()
        if value is None:
            value = 0
        widget.setValue(value)

    def setModelData(self, widget):
        self.prop.setValue(widget.value())

class _PropertyFloatEditor(_PropertyEditor):
    '''Editor for float values - uses a double spin box.'''

    def widget(self, parent):
        return QtGui.QDoubleSpinBox(parent)

    def setEditorData(self, widget):
        value = self.prop.getValue()
        if value is None:
            value = 0.0
        widget.setValue(value)

    def setModelData(self, widget):
        self.prop.setValue(widget.value())

class _PropertyFileEditor(_PropertyEditor):

    def widget(self, parent):
        return QtGui.QLineEdit(parent)

    def setEditorData(self, widget):
        text = '' if self.prop.getValue() is None else self.prop.getValue()
        widget.setText(text)

    def setModelData(self, widget):
        self.prop.setValue(widget.text())

_EditorFactory = {
        PathSetupSheetOpPrototype.Property:             None,
        PathSetupSheetOpPrototype.PropertyAngle:        _PropertyAngleEditor,
        PathSetupSheetOpPrototype.PropertyBool:         _PropertyBoolEditor,
        PathSetupSheetOpPrototype.PropertyDistance:     _PropertyLengthEditor,
        PathSetupSheetOpPrototype.PropertyEnumeration:  _PropertyEnumEditor,
        PathSetupSheetOpPrototype.PropertyFloat:        _PropertyFloatEditor,
        PathSetupSheetOpPrototype.PropertyInteger:      _PropertyIntegerEditor,
        PathSetupSheetOpPrototype.PropertyLength:       _PropertyLengthEditor,
        PathSetupSheetOpPrototype.PropertyPercent:      _PropertyPercentEditor,
        PathSetupSheetOpPrototype.PropertyString:       _PropertyStringEditor,
        }

def Editor(prop):
    '''Returns an editor class to be used for the given property.'''
    factory = _EditorFactory[prop.__class__]
    if factory:
        return factory(prop)
    return None



# -*- coding: utf-8 -*-

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import PathScripts.PathIconViewProvider as PathIconViewProvider
import PathScripts.PathLog as PathLog
import PathScripts.PathPreferences as PathPreferences
import PathScripts.PathToolBit as PathToolBit
import PathScripts.PathToolBitEdit as PathToolBitEdit
import os

from PySide import QtCore, QtGui

__title__ = "Tool Bit UI"
__author__ = "sliptonic (Brad Collette)"
__url__ = "http://www.freecadweb.org"
__doc__ = "Task panel editor for a ToolBit"

# Qt translation handling
def translate(context, text, disambig=None):
    return QtCore.QCoreApplication.translate(context, text, disambig)

#PathLog.setLevel(PathLog.Level.DEBUG, PathLog.thisModule())
#PathLog.trackModule(PathLog.thisModule())

class ViewProvider(object):
    '''ViewProvider for a ToolBit.
    It's sole job is to provide an icon and invoke the TaskPanel on edit.'''

    def __init__(self, vobj, name):
        PathLog.track(name, vobj.Object)
        self.panel = None
        self.icon = name
        self.obj = vobj.Object
        self.vobj = vobj
        vobj.Proxy = self

    def attach(self, vobj):
        PathLog.track(vobj.Object)
        self.vobj = vobj
        self.obj  = vobj.Object

    def getIcon(self):
        png = self.obj.Proxy.getBitThumbnail(self.obj)
        if png:
            pixmap = QtGui.QPixmap()
            pixmap.loadFromData(png, 'PNG')
            return QtGui.QIcon(pixmap)
        return ':/icons/Path-ToolBit.svg'

    def __getstate__(self):
        return None

    def __setstate__(self, state):
        # pylint: disable=unused-argument
        return None

    def onDelete(self, vobj, arg2=None):
        PathLog.track(vobj.Object.Label)
        vobj.Object.Proxy.onDelete(vobj.Object)

    def getDisplayMode(self, mode):
        # pylint: disable=unused-argument
        return 'Default'

    def _openTaskPanel(self, vobj, deleteOnReject):
        PathLog.track()
        self.panel = TaskPanel(vobj, deleteOnReject)
        FreeCADGui.Control.closeDialog()
        FreeCADGui.Control.showDialog(self.panel)
        self.panel.setupUi()

    def setCreate(self, vobj):
        PathLog.track()
        self._openTaskPanel(vobj, True)

    def setEdit(self, vobj, mode=0):
        # pylint: disable=unused-argument
        self._openTaskPanel(vobj, False)
        return True

    def unsetEdit(self, vobj, mode):
        # pylint: disable=unused-argument
        FreeCADGui.Control.closeDialog()
        self.panel = None
        return

    def claimChildren(self):
        if self.obj.BitBody:
            return [self.obj.BitBody]
        return []

    def doubleClicked(self, vobj):
        self.setEdit(vobj)

class TaskPanel:
    '''TaskPanel for the SetupSheet - if it is being edited directly.'''

    def __init__(self, vobj, deleteOnReject):
        PathLog.track(vobj.Object.Label)
        self.vobj = vobj
        self.obj = vobj.Object
        self.editor = PathToolBitEdit.ToolBitEditor(self.obj)
        self.form = self.editor.form
        self.deleteOnReject = deleteOnReject
        FreeCAD.ActiveDocument.openTransaction(translate('PathToolBit', 'Edit ToolBit'))

    def reject(self):
        FreeCAD.ActiveDocument.abortTransaction()
        self.editor.reject()
        FreeCADGui.Control.closeDialog()
        if self.deleteOnReject:
            FreeCAD.ActiveDocument.openTransaction(translate('PathToolBit', 'Uncreate ToolBit'))
            self.editor.reject()
            FreeCAD.ActiveDocument.removeObject(self.obj.Name)
            FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()

    def accept(self):
        self.editor.accept()

        FreeCAD.ActiveDocument.commitTransaction()
        FreeCADGui.ActiveDocument.resetEdit()
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()

    def updateUI(self):
        self.editor.updateUI()

    def updateModel(self):
        self.editor.updateTool()
        FreeCAD.ActiveDocument.recompute()

    def setupUi(self):
        self.editor.setupUI()


class ToolBitSelector(object):
    ToolRole = QtCore.Qt.UserRole + 1

    def __init__(self):
        self.buttons = None
        self.editor = None
        self.dialog = None
        self.form = FreeCADGui.PySideUic.loadUi(':/panels/ToolBitSelector.ui')
        self.setupUI()

    def updateTools(self, selected=None):
        PathLog.track()
        selItem = None
        self.form.tools.setUpdatesEnabled(False)
        if selected is None and self.form.tools.currentItem():
            selected = self.form.tools.currentItem().text()
        self.form.tools.clear()
        for tool in sorted(self.loadedTools(), key=lambda t: t.Label):
            icon = None
            if tool.ViewObject and tool.ViewObject.Proxy:
                icon = tool.ViewObject.Proxy.getIcon()
            if icon and isinstance(icon, QtGui.QIcon):
                item = QtGui.QListWidgetItem(icon, tool.Label)
            else:
                item = QtGui.QListWidgetItem(tool.Label)
            item.setData(self.ToolRole, tool)
            if selected == tool.Label:
                selItem = item
            self.form.tools.addItem(item)
        if selItem:
            self.form.tools.setCurrentItem(selItem)
        self.updateSelection()
        self.form.tools.setUpdatesEnabled(True)

    def getTool(self):
        PathLog.track()
        self.updateTools()
        res = self.form.exec_()
        if 1 == res and self.form.tools.currentItem():
            return self.form.tools.currentItem().data(self.ToolRole)
        return None

    def loadedTools(self):
        PathLog.track()
        if FreeCAD.ActiveDocument:
            return [o for o in FreeCAD.ActiveDocument.Objects if hasattr(o, 'Proxy') and isinstance(o.Proxy, PathToolBit.ToolBit)]
        return []

    def loadTool(self):
        PathLog.track()
        tool = LoadTool(self.form)
        if tool:
            self.updateTools(tool.Label)

    def createTool(self):
        PathLog.track()
        tool = PathToolBit.Factory.Create()

        def accept():
            self.editor.accept()
            self.dialog.done(1)
            self.updateTools(tool.Label)

        def reject():
            FreeCAD.ActiveDocument.openTransaction(translate('PathToolBit', 'Uncreate ToolBit'))
            self.editor.reject()
            self.dialog.done(0)
            FreeCAD.ActiveDocument.removeObject(tool.Name)
            FreeCAD.ActiveDocument.commitTransaction()

        self.dialog = QtGui.QDialog(self.form)
        layout = QtGui.QVBoxLayout(self.dialog)
        self.editor = PathToolBitEdit.ToolBitEditor(tool, self.dialog)
        self.editor.setupUI()
        self.buttons = QtGui.QDialogButtonBox(
                QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Cancel,
                QtCore.Qt.Horizontal, self.dialog)
        layout.addWidget(self.buttons)
        self.buttons.accepted.connect(accept)
        self.buttons.rejected.connect(reject)
        print(self.dialog.exec_())

    def updateSelection(self):
        PathLog.track()
        if self.form.tools.selectedItems():
            self.form.buttonBox.button(QtGui.QDialogButtonBox.Ok).setEnabled(True)
        else:
            self.form.buttonBox.button(QtGui.QDialogButtonBox.Ok).setEnabled(False)

    def setupUI(self):
        PathLog.track()
        self.form.toolCreate.clicked.connect(self.createTool)
        self.form.toolLoad.clicked.connect(self.loadTool)
        self.form.tools.itemSelectionChanged.connect(self.updateSelection)
        self.form.tools.doubleClicked.connect(self.form.accept)

class ToolBitGuiFactory(PathToolBit.ToolBitFactory):

    def Create(self, name='ToolBit', shapeFile=None):
        '''Create(name = 'ToolBit') ... creates a new tool bit.
        It is assumed the tool will be edited immediately so the internal bit body is still attached.'''
        FreeCAD.ActiveDocument.openTransaction(translate('PathToolBit', 'Create ToolBit'))
        tool = PathToolBit.ToolBitFactory.Create(self, name, shapeFile)
        PathIconViewProvider.Attach(tool.ViewObject, name)
        FreeCAD.ActiveDocument.commitTransaction()
        return tool

def GetToolFile(parent = None):
    if parent is None:
        parent = QtGui.QApplication.activeWindow()
    foo = QtGui.QFileDialog.getOpenFileName(parent, 'Tool', PathPreferences.lastPathToolBit(), '*.fctb')
    if foo and foo[0]:
        PathPreferences.setLastPathToolBit(os.path.dirname(foo[0]))
        return foo[0]
    return None

def GetToolFiles(parent = None):
    if parent is None:
        parent = QtGui.QApplication.activeWindow()
    foo = QtGui.QFileDialog.getOpenFileNames(parent, 'Tool', PathPreferences.lastPathToolBit(), '*.fctb')
    if foo and foo[0]:
        PathPreferences.setLastPathToolBit(os.path.dirname(foo[0][0]))
        return foo[0]
    return []


def LoadTool(parent = None):
    '''LoadTool(parent=None) ... Open a file dialog to load a tool from a file.'''
    foo = GetToolFile(parent)
    return PathToolBit.Factory.CreateFrom(foo) if foo else foo

def LoadTools(parent = None):
    '''LoadTool(parent=None) ... Open a file dialog to load a tool from a file.'''
    return [PathToolBit.Factory.CreateFrom(foo) for foo in GetToolFiles(parent)]

# Set the factory so all tools are created with UI
PathToolBit.Factory = ToolBitGuiFactory()

PathIconViewProvider.RegisterViewProvider('ToolBit', ViewProvider)



#/***************************************************************************
# *   Copyright (c) Victor Titov (DeepSOIC)                                 *
# *                                           (vv.titov@gmail.com) 2016     *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************/

# module is named mTempoVis, because Show.TimpoVis exposes the class as its member, and hides the module TempoVis.py.

from . import Containers

from . import TVStack

import FreeCAD as App
if App.GuiUp:
    import FreeCADGui as Gui
Wrn = lambda msg: App.Console.PrintWarning(msg + "\n")
Err = lambda msg: App.Console.PrintError(msg + "\n")
Log = lambda msg: App.Console.PrintLog(msg + "\n")

from copy import copy

S_EMPTY = 0 # TV is initialized, but no changes were done through it
S_ACTIVE = 1 # TV has something to be undone 
S_RESTORED = 2 # TV has been restored
S_INTERNAL = 3 # TV instance is being used by another TV instance as a redo data storage

def _printTraceback(err):
    import sys
    if err is sys.exc_info()[1]:
        import traceback
        tb = traceback.format_exc()
        Log(tb)

class MAINSTACK(object):
    '''it's just a default value definition for TV constructor'''
    pass
class JUST_SAVE(object):
    '''it's just a default value meaning "save current scene value but don't modify anything"'''
    pass

class TempoVis(object):
    '''TempoVis - helper object to save visibilities of objects before doing
    some GUI editing, hiding or showing relevant stuff during edit, and
    then restoring all visibilities after editing.

    Constructors:
    TempoVis(document, stack = MAINSTACK, **kwargs): creates a new TempoVis. 
    
    document: required. Objects not belonging to the document can't be modified via TempoVis.
    
    stack: optional. Which stack to insert this new TV into. Can be:
    a TVStack instance (then, the new TV is added to the top of the stack), 
    MAINSTACK special value (a global stack for the document will be used), or 
    None (then, the TV is not in any stack, and can be manually instertd into one if desired).
    
    Any additional keyword args are assigned as attributes. You can use it to immediately set a tag, for example.'''

    document = None
    stack = None # reference to stack this TV is in

    data = None # dict. key = ("class_id","key"), value = instance of SceneDetail
    data_requested = None #same as data, but stores (wanted) values passed to modify()
    
    state = S_EMPTY
    
    tag = '' #stores any user-defined string for identification purposes

    def _init_attrs(self):
        '''initialize member variables to empty values (needed because we can't use mutable initial values when initializing member variables in class definition)'''
        self.data = {}
        self.data_requested = {}
        
  #<core interface>
    def __init__(self, document, stack = MAINSTACK, **kwargs):
        self._init_attrs()
        self.document = document
        
        if stack is MAINSTACK:
            stack = TVStack.mainStack(document)
        
        if stack is None:
            pass
        else:
            stack.insert(self)
        
        for key,val in kwargs.items():
            setattr(self, key, val)
    
    def __del__(self):
        if self.state == S_ACTIVE:
            self.restore(ultimate= True)
    
    def has(self, detail):
        '''has(self, detail): returns True if this TV has this detail value saved.
        example: tv.has(VProperty(obj, "Visibility"))'''
        return detail.full_key in self.data
    
    def stored_val(self, detail):
        '''stored_val(self, detail): returns value of detail remembered by this TV. If not, raises KeyError.'''
        return self.data[detail.full_key].data

    def save(self, detail, mild_restore = False):
        '''save(detail, mild_restore = False):saves the scene detail to be restored. 
        The detail is saved only once; repeated calls are ignored.
        mild_restore: internal, do not use.'''
        self._change()
        if not detail.full_key in self.data:
            #not saved yet
            tv1, curr = self._value_after(detail, query_scene= True)
            self.data[detail.full_key] = copy(curr)
            self.data[detail.full_key].mild_restore = mild_restore
        else:
            #saved already. Change restore policy, if necessary.
            stored_dt = self.data[detail.full_key]
            if not mild_restore:
                stored_dt.mild_restore = False

    def modify(self, detail, mild_restore = None):
        '''modify(detail, mild_restore = True): modifies scene detail through this TV. 
        The value is provided as an instance of SceneDetail implementation.
        The procedure takes care to account for the stack - that is, if in a TV applied 
        later than this one this detail was changed too, the value saved therein is altered, 
        rather than applied to the scene.
        
        mild_restore: if True, when restoring later, checks if the value was changed 
        by user after last call to modify(), and doesn't restore if it was changed.
        
        Example: tv.modify(VProperty(obj, "Visibility", True))'''

        self._change()
        
        if mild_restore is not None:
            detail.mild_restore = mild_restore
        
        # save current
        self.save(detail, detail.mild_restore)
        
        # apply
        tv1, curr = self._value_after(detail)
        if tv1 is not None:
            tv1.data[detail.full_key].data = detail.data
        else:
            detail.apply_data(detail.data)
        
        # and record.
        if detail.mild_restore:
            self.data_requested[detail.full_key] = copy(detail)
        
    def restoreDetail(self, detail, ultimate = False):
        '''restoreDetail(detail, ultimate = False): restores a specific scene detail.
        ultimate: if true, the saved value is cleaned out.
        If the detail is not found, nothing is done.
        '''
        if not self.has(detail):
            return
        self._restore_detail(detail)
        if ultimate:
            self.forgetDetail(detail)
    
    
    def forgetDetail(self, detail):
        '''forgetDetail(detail): ditches a saved detail value, making the change done through this TV permanent.'''
        self.data.pop(detail.full_key, None)
        self.data_requested.pop(detail.full_key, None)
    
    def forget(self):
        '''forget(self): clears this TV, making all changes done through it permanent. 
        Also, withdraws the TV from the stack.'''
        self.state = S_EMPTY
        self.data = {}
        if self.is_in_stack:
            self.stack.withdraw(self)

    def restore(self, ultimate = True):
        '''restore(ultimate = True): undoes all changes done through this tempovis / restores saved scene details.
        ultimate: if true, the saved values are cleaned out, and the TV is withdrawn from 
        the stack. If false, the TV will still remember stuff, and restore can be called again.
        '''
        if self.state == S_RESTORED:
            return

        if self.state != S_INTERNAL and ultimate:
            self.state = S_RESTORED        
        
        for key, detail in self.data.items():
            try:
                self._restoreDetail(detail)
            except Exception as err:
                Err("TempoVis.restore: failed to restore detail {key}: {err}".format(key= key, err= str(err)))
                _printTraceback(err)
        if ultimate:
            self.data = {}
            if self.is_in_stack:
                self.stack.withdraw(self)
    
  #</core interface>
  
  #<stack interface>
    def _inserted(self, stack, index):
        '''calles when this tv is inserted into a stack'''
        self.stack = stack
    def _withdrawn(self, stack, index):
        '''calles when this tv is withdrawn from a stack'''
        self.stack = None
    @property
    def is_in_stack(self):
        return self.stack is not None
  #</stack interface>

  #<convenience functions>            
    def modifyVPProperty(self, doc_obj_or_list, prop_names, new_value = JUST_SAVE, mild_restore = None):
        '''modifyVPProperty(doc_obj_or_list, prop_names, new_value = JUST_SAVE, mild_restore = None): modifies
        prop_name property of ViewProvider of doc_obj_or_list, and remembers
        original value of the property. Original values will be restored upon
        TempoVis deletion, or call to restore().
        
        mild_restore: test if user changed the value manually when restoring the TV.'''
        
        if self.state == S_RESTORED:
            Wrn("Attempting to use a TV that has been restored. There must be a problem with code.")
            return

        if not hasattr(doc_obj_or_list, '__iter__'):
            doc_obj_or_list = [doc_obj_or_list]
        if not isinstance(prop_names,(list,tuple)):
            prop_names = [prop_names]
        for doc_obj in doc_obj_or_list:
            for prop_name in prop_names:
                if not hasattr(doc_obj.ViewObject, prop_name):
                    Wrn("TempoVis: object {obj} has no attribute {attr}. Skipped."
                                             .format(obj= doc_obj.Name, attr= prop_name))
                    continue

                # Because the introduction of external objects, we shall now
                # accept objects from all opened documents.
                #
                #  if doc_obj.Document is not self.document:  #ignore objects from other documents
                #      raise ValueError("Document object to be modified does not belong to document TempoVis was made for.")
                from .SceneDetails.VProperty import VProperty
                if new_value is JUST_SAVE:
                    if mild_restore:
                        Wrn("TempoVis: can't just save a value for mild restore. Saving for hard restore.")
                    self.save(VProperty(doc_obj, prop_name, new_value))
                else:
                    self.modify(VProperty(doc_obj, prop_name, new_value), mild_restore)

    def restoreVPProperty(self, doc_obj_or_list, prop_names):
        '''restoreVPProperty(doc_obj_or_list, prop_name, new_value): restores specific property changes.'''
        from .SceneDetails.VProperty import VProperty
        
        if not hasattr(doc_obj_or_list, '__iter__'):
            doc_obj_or_list = [doc_obj_or_list]
        if not isinstance(prop_names,(tuple,list)):
            prop_names = [prop_names]
        for doc_obj in doc_obj_or_list:
            for prop_name in prop_names:
                try:
                    self.restoreDetail(VProperty(doc_obj, prop_name))
                except Exception as err:
                    Err("TempoVis.restore: failed to restore detail {key}: {err}".format(key= key, err= str(err)))
                    _printTraceback(err)


    def saveBodyVisibleFeature(self, doc_obj_or_list):
        """saveBodyVisibleFeature(self, doc_obj_or_list): saves Visibility of currently 
        visible feature, for every body of PartDesign features in the provided list."""
        if not hasattr(doc_obj_or_list, '__iter__'):
            doc_obj_or_list = [doc_obj_or_list]
        objs = []
        bodies = set()
        for obj in doc_obj_or_list:
            body = getattr(obj,'_Body',None)
            if not body or body in bodies:
                continue
            bodies.add(body)
            feature = getattr(body,'VisibleFeature',None)
            if feature:
                objs.append(feature)
        self.modifyVPProperty(objs, 'Visibility', JUST_SAVE)
        return objs

    def show(self, doc_obj_or_list, links_too = True, mild_restore = None):
        '''show(doc_obj_or_list, links_too = True): shows objects (sets their Visibility to True). 
        doc_obj_or_list can be a document object, or a list of document objects.
        If links_too is True, all Links of the objects are also hidden, by setting LinkVisibility attribute of each object.'''
        doc_obj_or_list = self._3D_objects(doc_obj_or_list)
        self.saveBodyVisibleFeature(doc_obj_or_list) #fix implicit hiding of other features by PartDesign not being recorded to TV
        self.modifyVPProperty(doc_obj_or_list, 'Visibility', True, mild_restore)
        if links_too:
            self.modifyVPProperty(doc_obj_or_list, 'LinkVisibility', True, mild_restore)

    def hide(self, doc_obj_or_list, links_too = True, mild_restore = None):
        '''hide(doc_obj_or_list): hides objects (sets their Visibility to False). doc_obj_or_list can be a document object, or a list of document objects'''
        doc_obj_or_list = self._3D_objects(doc_obj_or_list)
        # no need to saveBodyVisibleFeature here, as no implicit showing will happen
        self.modifyVPProperty(doc_obj_or_list, 'Visibility', False, mild_restore)
        if links_too:
            self.modifyVPProperty(doc_obj_or_list, 'LinkVisibility', False, mild_restore)

    def get_all_dependent(self, doc_obj, subname = None):
        '''get_all_dependent(doc_obj, subname = None): gets all objects that depend on doc_obj. Containers and Links (if subname) required for visibility of the object are excluded from the list.'''
        from . import Containers
        from .Containers import isAContainer
        from .DepGraphTools import getAllDependencies, getAllDependent
        if subname:
            # a link-path was provided. doc_obj has nothing to do with the object we want 
            # to collect dependencies from. So, replace it with the one pointed by link-path.
            cnt_chain = doc_obj.getSubObjectList(subname)
            doc_obj = cnt_chain[-1].getLinkedObject()
            # cnt_chain can either end with the object (e.g. if a sketch is in a part, and 
            # a link is to a part), or it may be a Link object (if we have a straight or 
            # even nested Link to the sketch).
            #
            # I don't know why do we need that isAContainer check here, but I'm leaving it, 
            # realthunder must be knowing his business --DeepSOIC
            cnt_chain = [ o for o in cnt_chain
                            if o==cnt_chain[-1] or isAContainer(o, links_too= True) ]
        else:
            cnt_chain = Containers.ContainerChain(doc_obj)
        return [o for o in getAllDependent(doc_obj) if not o in cnt_chain]

    def hide_all_dependent(self, doc_obj):
        '''hide_all_dependent(doc_obj): hides all objects that depend on doc_obj. Groups, Parts and Bodies are not hidden by this.'''
        self.hide(self._3D_objects(self.get_all_dependent(doc_obj)))

    def show_all_dependent(self, doc_obj):
        '''show_all_dependent(doc_obj): shows all objects that depend on doc_obj. This method is probably useless.'''
        from .DepGraphTools import getAllDependencies, getAllDependent
        self.show(self._3D_objects(getAllDependent(doc_obj)))

    def restore_all_dependent(self, doc_obj):
        '''show_all_dependent(doc_obj): restores original visibilities of all dependent objects.'''
        from .DepGraphTools import getAllDependencies, getAllDependent
        self.restoreVPProperty( getAllDependent(doc_obj), ('Visibility', 'LinkVisibility') )

    def hide_all_dependencies(self, doc_obj):
        '''hide_all_dependencies(doc_obj): hides all objects that doc_obj depends on (directly and indirectly).'''
        from .DepGraphTools import getAllDependencies, getAllDependent
        self.hide(self._3D_objects(getAllDependencies(doc_obj)))

    def show_all_dependencies(self, doc_obj):
        '''show_all_dependencies(doc_obj): shows all objects that doc_obj depends on (directly and indirectly). This method is probably useless.'''
        from .DepGraphTools import getAllDependencies, getAllDependent
        self.show(self._3D_objects(getAllDependencies(doc_obj)))

    def saveCamera(self, vw = None):
        self._change()
        from .SceneDetails.Camera import Camera
        self.save(Camera(self.document))
        
    def restoreCamera(self, ultimate = False):
        from .SceneDetails.Camera import Camera
        dt = Camera(self.document)
        self.restoreDetail(dt, ultimate)

    def setUnpickable(self, doc_obj_or_list, actual_pick_style = 2): #2 is coin.SoPickStyle.UNPICKABLE
        '''setUnpickable(doc_obj_or_list, actual_pick_style = 2): sets object unpickable (transparent to clicks).
        doc_obj_or_list: object or list of objects to alter (App)
        actual_pick_style: optional parameter, specifying the actual pick style: 
        0 = regular, 1 = bounding box, 2 (default) = unpickable.
        
        Implementation detail: uses SoPickStyle node. If viewprovider already has a node 
        of this type as direct child, one is used. Otherwise, new one is created and 
        inserted as the very first node, and remains there even after restore()/deleting 
        tempovis. '''
        
        from .SceneDetails.Pickability import Pickability
        from .ShowUtils import is3DObject
        
        if not hasattr(doc_obj_or_list, '__iter__'):
            doc_obj_or_list = [doc_obj_or_list]
        for doc_obj in doc_obj_or_list:
            if not is3DObject(doc_obj):
                continue
            dt = Pickability(doc_obj, actual_pick_style)
            self.modify(dt)

    def clipPlane(self, doc_obj_or_list, enable, placement, offset = 0.02): 
        '''clipPlane(doc_obj_or_list, enable, placement, offset): slices off the object with a clipping plane.
        doc_obj_or_list: object or list of objects to alter (App)
        enable: True if you want clipping, False if you want to remove clipping: 
        placement: XY plane of local coordinates of the placement is the clipping plane. The placement must be in document's global coordinate system.
        offset: shifts the plane. Positive offset reveals more of the object.
        
        Implementation detail: uses SoClipPlane node. If viewprovider already has a node 
        of this type as direct child, one is used. Otherwise, new one is created and 
        inserted as the very first node. The node is left, but disabled when tempovis is restoring.'''
        
        from .SceneDetails.ObjectClipPlane import ObjectClipPlane
        from .ShowUtils import is3DObject
        
        if not hasattr(doc_obj_or_list, '__iter__'):
            doc_obj_or_list = [doc_obj_or_list]
        for doc_obj in doc_obj_or_list:
            if not is3DObject(doc_obj):
                continue
            dt = ObjectClipPlane(doc_obj, enable, placement, offset)
            self.modify(dt)
                    
    @staticmethod
    def allVisibleObjects(aroundObject):
        '''allVisibleObjects(aroundObject): returns list of objects that have to be toggled invisible for only aroundObject to remain. 
        If a whole container can be made invisible, it is returned, instead of its child objects.'''
        from .ShowUtils import is3DObject
        from . import Containers
        
        chain = Containers.VisGroupChain(aroundObject)
        result = []
        for i in range(len(chain)):
            cnt = chain[i]
            cnt_next = chain[i+1] if i+1 < len(chain) else aroundObject
            container = Containers.Container(cnt)
            for obj in container.getVisGroupChildren():
                if not is3DObject(obj):
                    continue
                if obj is not cnt_next:
                    if container.isChildVisible(obj):
                        result.append(obj)
        return result
        
    def sketchClipPlane(self, sketch, enable = None):
        '''sketchClipPlane(sketch, enable = None): Clips all objects by plane of sketch. 
        If enable argument is omitted, calling the routine repeatedly will toggle clipping plane.'''
        
        from .SceneDetails.ClipPlane import ClipPlane
        
        editDoc = Gui.editDocument()
        if editDoc is None:
            doc = sketch.Document
            pla = sketch.getGlobalPlacement()
        else:
            doc = editDoc.Document
            pla = App.Placement(editDoc.EditingTransform)
        toggle = {False: 0, True: 1, None: -1}[enable]
        
        self.modify(ClipPlane(doc, toggle, pla, 0.02))
    
    def activateWorkbench(self, wb_name):
        from .SceneDetails.Workbench import Workbench
        self.modify(Workbench(wb_name))
    
  #</convenience functions>

  #<internals>
    def _restoreDetail(self, detail):
        p = self.data[detail.full_key]
        tv1, curr = self._value_after(detail, query_scene= p.mild_restore)
        if p.mild_restore:
            if self.data_requested[detail.full_key] != curr:
                #the value on the scene doesn't match what was requested through TV. User probably changed it. We don't want to mess it up.
                self._purge_milds(detail)
                return
        if tv1 is None:
            # no other TV has changed this detail later, apply to the scene
            detail.apply_data(p.data)
        else:
            #modify saved detail of higher TV
            tv1.data[detail.full_key].data = p.data
    
    def _purge_milds(self, detail):
        """_purge_milds(detail): wipes out detail from earlier TVs if the detail is mild-restore."""
        if not self.is_in_stack:
            return
        seq_before, seq_after = self.stack.getSplitSequence(self)
        for tv in reversed(seq_before):
            if tv.has(detail):
                if tv.data[detail.full_key].mild_restore:
                    tv.forgetDetail(detail)
                else:
                    #hard-restoring value encountered, stop
                    break

    def _change(self):
        '''to be called whenever anything is done that is to be restored later.'''
        if self.state == S_EMPTY:
            self.state = S_ACTIVE
        if self.state == S_RESTORED:
            Wrn("Attempting to use a TV that has been restored. There must be a problem with code.")
        self.tv_redo = None    

    def _value_after(self, detail, query_scene = False):
        '''_value_current(detail): returns (tv, detail1). SceneDetail instance holds "current" value of 
        scene detail (current from the context of this TV; i.e. either the current scene 
        status, or the saved state from upper TVs).
        If no upper TV has saved the detail value, returns either (None, None), or 
        (None, detail1) if query_scene is True, where detail1 holds value from the scene.'''
        def scene_value():
            if query_scene:
                cpy = copy(detail)
                cpy.data = cpy.scene_value()
                return (None, cpy)
            else:
                return (None, None)
            
        if self.is_in_stack:
            va = self.stack.value_after(self, detail)
            if va is None:
                return scene_value()
            else:
                return va
        else:
            return scene_value()
            
    def _3D_objects(self, doc_obj_or_list):
        """_3D_objects(doc_obj_or_list): returns list of objects that are in 3d view."""
        from .ShowUtils import is3DObject
        
        if not hasattr(doc_obj_or_list, '__iter__'):
            doc_obj_or_list = [doc_obj_or_list]
        
        return [obj for obj in doc_obj_or_list if is3DObject(obj)]
        
    def __getstate__(self):
        return None
        
    def __setstate__(self, state):
        self._init_attrs()
    
        



# -*- coding: utf-8 -*-

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

'''
The purpose of this file is to collect some handy functions. The reason they
are not in PathUtils (and there is this confusing naming going on) is that
PathUtils depends on PathJob. Which makes it impossible to use the functions
and classes defined there in PathJob.

So if you add to this file and think about importing anything from PathScripts
other than PathLog, then it probably doesn't belong here.
'''

import six
import PathScripts.PathLog as PathLog
import PySide

PathLog.setLevel(PathLog.Level.INFO, PathLog.thisModule())

def translate(context, text, disambig=None):
    return PySide.QtCore.QCoreApplication.translate(context, text, disambig)

def _getProperty(obj, prop):
    o = obj
    attr = obj
    name = None
    for name in prop.split('.'):
        o = attr
        if not hasattr(o, name):
            break
        attr = getattr(o, name)

    if o == attr:
        PathLog.warning(translate('PathGui', "%s has no property %s (%s))") % (obj.Label, prop, name))
        return (None, None, None)

    #PathLog.debug("found property %s of %s (%s: %s)" % (prop, obj.Label, name, attr))
    return(o, attr, name)

def getProperty(obj, prop):
    '''getProperty(obj, prop) ... answer obj's property defined by its canonical name.'''
    o, attr, name = _getProperty(obj, prop) # pylint: disable=unused-variable
    return attr

def getPropertyValueString(obj, prop):
    '''getPropertyValueString(obj, prop) ... answer a string representation of an object's property's value.'''
    attr = getProperty(obj, prop)
    if hasattr(attr, 'UserString'):
        return attr.UserString
    return str(attr)

def setProperty(obj, prop, value):
    '''setProperty(obj, prop, value) ... set the property value of obj's property defined by its canonical name.'''
    o, attr, name = _getProperty(obj, prop) # pylint: disable=unused-variable
    if o and name:
        setattr(o, name, value)

# NotValidBaseTypeIds = ['Sketcher::SketchObject']
NotValidBaseTypeIds = []


def isValidBaseObject(obj):
    '''isValidBaseObject(obj) ... returns true if the object can be used as a base for a job.'''
    if hasattr(obj, 'getParentGeoFeatureGroup') and obj.getParentGeoFeatureGroup():
        # Can't link to anything inside a geo feature group anymore
        PathLog.debug("%s is inside a geo feature group" % obj.Label)
        return False
    if hasattr(obj, 'BitBody') and hasattr(obj, 'BitShape'):
        # ToolBit's are not valid base objects
        return False
    if obj.TypeId in NotValidBaseTypeIds:
        PathLog.debug("%s is blacklisted (%s)" % (obj.Label, obj.TypeId))
        return False
    if hasattr(obj, 'Sheets') or hasattr(obj, 'TagText'): # Arch.Panels and Arch.PanelCut
        PathLog.debug("%s is not an Arch.Panel" % (obj.Label))
        return False
    import Part
    return not Part.getShape(obj).isNull()

def isSolid(obj):
    '''isSolid(obj) ... return True if the object is a valid solid.'''
    import Part
    shape = Part.getShape(obj)
    return not shape.isNull() and shape.Volume and shape.isClosed()

def opProperty(op, prop):
    '''opProperty(op, prop) ... return the value of property prop of the underlying operation (or None if prop does not exist)'''
    if hasattr(op, prop):
        return getattr(op, prop)
    if hasattr(op, 'Base'):
        return opProperty(op.Base, prop)
    return None

def toolControllerForOp(op):
    '''toolControllerForOp(op) ... return the tool controller used by the op.
    If the op doesn't have its own tool controller but has a Base object, return its tool controller.
    Otherwise return None.'''
    return opProperty(op, 'ToolController')

def getPublicObject(obj):
    '''getPublicObject(obj) ... returns the object which should be used to reference a feature of the given object.'''
    if hasattr(obj, 'getParentGeoFeatureGroup'):
        body = obj.getParentGeoFeatureGroup()
        if body:
            return getPublicObject(body)
    return obj

def clearExpressionEngine(obj):
    '''clearExpressionEngine(obj) ... removes all expressions from obj.

There is currently a bug that invalidates the DAG if an object
is deleted that still has one or more expressions attached to it.
Use this function to remove all expressions before deletion.'''
    if hasattr(obj, 'ExpressionEngine'):
        for attr, expr in obj.ExpressionEngine: # pylint: disable=unused-variable
            obj.setExpression(attr, None)

def toUnicode(string):
    '''toUnicode(string) ... returns a unicode version of string regardless of the python version.'''
    return six.text_type(string)

def isString(string):
    '''isString(string) ... return True if string is a string, regardless of string type and python version.'''
    return isinstance(string, six.string_types)

def keyValueIter(dictionary):
    '''keyValueIter(dict) ... return iterable object over dictionary's (key,value) tuples.'''
    return six.iteritems(dictionary)



# -*- coding: utf-8 -*-

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2018 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
# pylint: disable=unused-import

import PathScripts.PathLog as PathLog

LOGLEVEL = False

if LOGLEVEL:
    PathLog.setLevel(PathLog.Level.DEBUG, PathLog.thisModule())
    PathLog.trackModule(PathLog.thisModule())
else:
    PathLog.setLevel(PathLog.Level.INFO, PathLog.thisModule())

Processed = False


def Startup():
    global Processed # pylint: disable=global-statement
    if not Processed:
        PathLog.debug('Initializing PathGui')
        from PathScripts import PathAdaptiveGui
        from PathScripts import PathArray
        from PathScripts import PathComment
        from PathScripts import PathCustom
        from PathScripts import PathDeburrGui
        from PathScripts import PathDressupAxisMap
        from PathScripts import PathDressupDogbone
        from PathScripts import PathDressupDragknife
        from PathScripts import PathDressupRampEntry
        from PathScripts import PathDressupPathBoundaryGui
        from PathScripts import PathDressupTagGui
        from PathScripts import PathDressupLeadInOut
        from PathScripts import PathDressupZCorrect
        from PathScripts import PathDrillingGui
        from PathScripts import PathEngraveGui
        from PathScripts import PathFixture
        from PathScripts import PathHelixGui
        from PathScripts import PathHop
        from PathScripts import PathInspect
        from PathScripts import PathMillFaceGui
        from PathScripts import PathPocketGui
        from PathScripts import PathPocketShapeGui
        from PathScripts import PathPost
        from PathScripts import PathProbeGui
        from PathScripts import PathProfileContourGui
        from PathScripts import PathProfileEdgesGui
        from PathScripts import PathProfileFacesGui
        from PathScripts import PathSanity
        from PathScripts import PathSetupSheetGui
        from PathScripts import PathSimpleCopy
        from PathScripts import PathSimulatorGui
        from PathScripts import PathStop
        # from PathScripts import PathSurfaceGui  # Added in initGui.py due to OCL dependency
        from PathScripts import PathToolController
        from PathScripts import PathToolControllerGui
        from PathScripts import PathToolLibraryManager
        from PathScripts import PathToolLibraryEditor
        from PathScripts import PathUtilsGui
        # from PathScripts import PathWaterlineGui  # Added in initGui.py due to OCL dependency
        Processed = True
    else:
        PathLog.debug('Skipping PathGui initialisation')



# ***************************************************************************
# *   Copyright (c) 2019 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

# to run the example use:
"""
from femexamples import ccx_cantilever_std as canti

canti.setup_cantileverbase()
canti.setup_cantileverfaceload()
canti.setup_cantilevernodeload()
canti.setup_cantileverprescribeddisplacement()
canti.setup_cantileverhexa20faceload()

"""

import FreeCAD

import Fem
import ObjectsFem

mesh_name = "Mesh"  # needs to be Mesh to work with unit tests


def init_doc(doc=None):
    if doc is None:
        doc = FreeCAD.newDocument()
    return doc


def setup_cantileverbase(doc=None, solvertype="ccxtools"):
    # setup CalculiX cantilever base model

    if doc is None:
        doc = init_doc()

    # geometry object
    # name is important because the other method in this module use obj name
    geom_obj = doc.addObject("Part::Box", "Box")
    geom_obj.Height = geom_obj.Width = 1000
    geom_obj.Length = 8000
    doc.recompute()

    if FreeCAD.GuiUp:
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")

    # solver
    if solvertype == "calculix":
        solver_object = analysis.addObject(
            ObjectsFem.makeSolverCalculix(doc, "SolverCalculiX")
        )[0]
    elif solvertype == "ccxtools":
        solver_object = analysis.addObject(
            ObjectsFem.makeSolverCalculixCcxTools(doc, "CalculiXccxTools")
        )[0]
        solver_object.WorkingDir = u""
    elif solvertype == "elmer":
        analysis.addObject(ObjectsFem.makeSolverElmer(doc, "SolverElmer"))
    elif solvertype == "z88":
        analysis.addObject(ObjectsFem.makeSolverZ88(doc, "SolverZ88"))
    if solvertype == "calculix" or solvertype == "ccxtools":
        solver_object.SplitInputWriter = False
        solver_object.AnalysisType = "static"
        solver_object.GeometricalNonlinearity = "linear"
        solver_object.ThermoMechSteadyState = False
        solver_object.MatrixSolverType = "default"
        solver_object.IterationsControlParameterTimeUse = False

    # material
    material_object = analysis.addObject(
        ObjectsFem.makeMaterialSolid(doc, "FemMaterial")
    )[0]
    mat = material_object.Material
    mat["Name"] = "CalculiX-Steel"
    mat["YoungsModulus"] = "210000 MPa"
    mat["PoissonRatio"] = "0.30"
    mat["Density"] = "7900 kg/m^3"
    mat["ThermalExpansionCoefficient"] = "0.012 mm/m/K"
    material_object.Material = mat

    # fixed_constraint
    fixed_constraint = analysis.addObject(
        ObjectsFem.makeConstraintFixed(doc, name="ConstraintFixed")
    )[0]
    fixed_constraint.References = [(geom_obj, "Face1")]

    # mesh
    from .meshes.mesh_canticcx_tetra10 import create_nodes, create_elements
    fem_mesh = Fem.FemMesh()
    control = create_nodes(fem_mesh)
    if not control:
        FreeCAD.Console.PrintError("Error on creating nodes.\n")
    control = create_elements(fem_mesh)
    if not control:
        FreeCAD.Console.PrintError("Error on creating elements.\n")
    femmesh_obj = analysis.addObject(
        doc.addObject("Fem::FemMeshObject", mesh_name)
    )[0]
    femmesh_obj.FemMesh = fem_mesh

    doc.recompute()
    return doc


def setup_cantileverfaceload(doc=None, solvertype="ccxtools"):
    # setup CalculiX cantilever, apply 9 MN on surface of front end face

    doc = setup_cantileverbase(doc, solvertype)

    # force_constraint
    force_constraint = doc.Analysis.addObject(
        ObjectsFem.makeConstraintForce(doc, name="ConstraintForce")
    )[0]
    force_constraint.References = [(doc.Box, "Face2")]
    force_constraint.Force = 9000000.0
    force_constraint.Direction = (doc.Box, ["Edge5"])
    force_constraint.Reversed = True

    doc.recompute()
    return doc


def setup_cantilevernodeload(doc=None, solvertype="ccxtools"):
    # setup CalculiX cantilever, apply 9 MN on the 4 nodes of the front end face

    doc = setup_cantileverbase(doc, solvertype)

    # force_constraint
    force_constraint = doc.Analysis.addObject(
        ObjectsFem.makeConstraintForce(doc, name="ConstraintForce")
    )[0]
    # should be possible in one tuple too
    force_constraint.References = [
        (doc.Box, "Vertex5"),
        (doc.Box, "Vertex6"),
        (doc.Box, "Vertex7"),
        (doc.Box, "Vertex8")
    ]
    force_constraint.Force = 9000000.0
    force_constraint.Direction = (doc.Box, ["Edge5"])
    force_constraint.Reversed = True

    doc.recompute()
    return doc


def setup_cantileverprescribeddisplacement(doc=None, solvertype="ccxtools"):
    # setup CalculiX cantilever
    # apply a prescribed displacement of 250 mm in -z on the front end face

    doc = setup_cantileverbase(doc, solvertype)

    # displacement_constraint
    displacement_constraint = doc.Analysis.addObject(
        ObjectsFem.makeConstraintDisplacement(doc, name="ConstraintDisplacmentPrescribed")
    )[0]
    displacement_constraint.References = [(doc.Box, "Face2")]
    displacement_constraint.zFix = False
    displacement_constraint.zFree = False
    displacement_constraint.zDisplacement = -250.0

    doc.recompute()
    return doc


def setup_cantileverhexa20faceload(doc=None, solvertype="ccxtools"):
    doc = setup_cantileverfaceload(doc, solvertype)

    # load the hexa20 mesh
    from .meshes.mesh_canticcx_hexa20 import create_nodes, create_elements
    fem_mesh = Fem.FemMesh()
    control = create_nodes(fem_mesh)
    if not control:
        FreeCAD.Console.PrintError("Error on creating nodes.\n")
    control = create_elements(fem_mesh)
    if not control:
        FreeCAD.Console.PrintError("Error on creating elements.\n")

    # overwrite mesh with the hexa20 mesh
    doc.getObject(mesh_name).FemMesh = fem_mesh

    doc.recompute()
    return doc



# -*- coding: utf-8 -*-

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import PathScripts
import os

from PySide import QtCore

class CommandToolBitCreate:
    '''
    Command used to create a new Tool.
    '''

    def __init__(self):
        pass

    def GetResources(self):
        return {'Pixmap': 'Path-ToolBit',
                'MenuText': QtCore.QT_TRANSLATE_NOOP("PathToolBit", "Create Tool"),
                'ToolTip': QtCore.QT_TRANSLATE_NOOP("PathToolBit", "Creates a new ToolBit object")}

    def IsActive(self):
        return FreeCAD.ActiveDocument is not None

    def Activated(self):
        obj = PathScripts.PathToolBit.Factory.Create()
        obj.ViewObject.Proxy.setCreate(obj.ViewObject)

class CommandToolBitSave:
    '''
    Command used to save an existing Tool to a file.
    '''

    def __init__(self, saveAs):
        self.saveAs = saveAs

    def GetResources(self):
        if self.saveAs:
            menuTxt = QtCore.QT_TRANSLATE_NOOP("PathToolBit", "Save Tool as...")
        else:
            menuTxt = QtCore.QT_TRANSLATE_NOOP("PathToolBit", "Save Tool")
        return {'Pixmap': 'Path-ToolBit',
                'MenuText': menuTxt,
                'ToolTip': QtCore.QT_TRANSLATE_NOOP("PathToolBit", "Save an existing ToolBit object to a file")}

    def selectedTool(self):
        sel = FreeCADGui.Selection.getSelectionEx()
        if 1 == len(sel) and isinstance(sel[0].Object.Proxy, PathScripts.PathToolBit.ToolBit):
            return sel[0].Object
        return None

    def IsActive(self):
        tool = self.selectedTool()
        if tool:
            if tool.File:
                return True
            return self.saveAs
        return False

    def Activated(self):
        from PySide import QtGui
        tool = self.selectedTool()
        if tool:
            path = None
            if not tool.File or self.saveAs:
                if tool.File:
                    fname = tool.File
                else:
                    fname = os.path.join(PathScripts.PathPreferences.lastPathToolBit(), tool.Label + '.fctb')
                foo = QtGui.QFileDialog.getSaveFileName(QtGui.QApplication.activeWindow(), "Tool", fname, "*.fctb")
                if foo:
                    path = foo[0]
            else:
                path = tool.File

            if path:
                if not path.endswith('.fctb'):
                    path += '.fctb'
                tool.Proxy.saveToFile(tool, path)
                PathScripts.PathPreferences.setLastPathToolBit(os.path.dirname(path))

class CommandToolBitLoad:
    '''
    Command used to load an existing Tool from a file into the current document.
    '''

    def __init__(self):
        pass

    def GetResources(self):
        return {'Pixmap': 'Path-ToolBit',
                'MenuText': QtCore.QT_TRANSLATE_NOOP("PathToolBit", "Load Tool"),
                'ToolTip': QtCore.QT_TRANSLATE_NOOP("PathToolBit", "Load an existing ToolBit object from a file")}

    def selectedTool(self):
        sel = FreeCADGui.Selection.getSelectionEx()
        if 1 == len(sel) and isinstance(sel[0].Object.Proxy, PathScripts.PathToolBit.ToolBit):
            return sel[0].Object
        return None

    def IsActive(self):
        return FreeCAD.ActiveDocument is not None

    def Activated(self):
        if PathScripts.PathToolBitGui.LoadTools():
            FreeCAD.ActiveDocument.recompute()

if FreeCAD.GuiUp:
    FreeCADGui.addCommand('Path_ToolBitCreate', CommandToolBitCreate())
    FreeCADGui.addCommand('Path_ToolBitLoad',   CommandToolBitLoad())
    FreeCADGui.addCommand('Path_ToolBitSave',   CommandToolBitSave(False))
    FreeCADGui.addCommand('Path_ToolBitSaveAs', CommandToolBitSave(True))

CommandList = ['Path_ToolBitCreate', 'Path_ToolBitLoad', 'Path_ToolBitSave', 'Path_ToolBitSaveAs']

FreeCAD.Console.PrintLog("Loading PathToolBitCmd... done\n")



# -*- coding: utf8 -*-
# Import gui init module
# (c) 2003 Juergen Riegel
#
# Gathering all the information to start FreeCAD
# This is the second one of three init scripts, the third one
# runs when the gui is up

#***************************************************************************
#*   Copyright (c) 2002 Juergen Riegel <juergen.riegel@web.de>             *
#*                                                                         *
#*   This file is part of the FreeCAD CAx development system.              *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   FreeCAD is distributed in the hope that it will be useful,            *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Lesser General Public License for more details.                   *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with FreeCAD; if not, write to the Free Software        *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************/


# Registered in Part's Init.py file
FreeCAD.changeImportModule("STEP with colors (*.step *.stp)","Import","ImportGui")
FreeCAD.changeExportModule("STEP with colors (*.step *.stp)","Import","ImportGui")

"""
class ImportWorkbench ( Workbench ):
	"Import workbench object"
	def Activate(self):
		# load the module
		try:
			Log ('Loading ImportGui module')
			import Import
			import ImportGui
		except ImportError:
			Err('Cannot load ImportGui')
			raise
	def GetIcon(self):
		# returns an icon for the workbench
		return ["/* XPM */\n"
			"static const char *fileopen[] = {\n"
			"\"16 13 5 1\",\n"
			"\". c #040404\",\n"
			"\"# c #808304\",\n"
			"\"a c None\",\n"
			"\"b c #f3f704\",\n"
			"\"c c #f3f7f3\",\n"
			"\"aaaaaaaaa...aaaa\",\n"
			"\"aaaaaaaa.aaa.a.a\",\n"
			"\"aaaaaaaaaaaaa..a\",\n"
			"\"a...aaaaaaaa...a\",\n"
			"\".bcb.......aaaaa\",\n"
			"\".cbcbcbcbc.aaaaa\",\n"
			"\".bcbcbcbcb.aaaaa\",\n"
			"\".cbcb...........\",\n"
			"\".bcb.#########.a\",\n"
			"\".cb.#########.aa\",\n"
			"\".b.#########.aaa\",\n"
			"\"..#########.aaaa\",\n"
			"\"...........aaaaa\"};\n"]

Gui.addWorkbench("Import",ImportWorkbench())
"""
# See https://forum.freecadweb.org/viewtopic.php?f=3&t=26782
#import Import_rc
#FreeCADGui.addPreferencePage(":/ui/preferences-import.ui","Import-Export")



#***************************************************************************
#*   Copyright (c) 2020 Adam Spontarelli <adam@vector-space.org>           *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************

import FreeCAD, Part
from fcsprocket import fcsprocket
from fcsprocket import sprocket

if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtCore, QtGui
    from FreeCADGui import PySideUic as uic

__title__="PartDesign SprocketObject management"
__author__ = "Adam Spontarelli"
__url__ = "http://www.freecadweb.org"


def makeSprocket(name):
    """
    makeSprocket(name): makes a Sprocket
    """
    obj = FreeCAD.ActiveDocument.addObject("Part::Part2DObjectPython",name)
    Sprocket(obj)
    if FreeCAD.GuiUp:
        ViewProviderSprocket(obj.ViewObject)
    #FreeCAD.ActiveDocument.recompute()
    if FreeCAD.GuiUp:
        body=FreeCADGui.ActiveDocument.ActiveView.getActiveObject("pdbody")
        part=FreeCADGui.ActiveDocument.ActiveView.getActiveObject("part")
        if body:
            body.Group=body.Group+[obj]
        elif part:
            part.Group=part.Group+[obj]
    return obj

class CommandSprocket:

    """
    the Fem Sprocket command definition
    """

    def GetResources(self):
        return {'Pixmap'  : 'PartDesign_Sprocket',
                'MenuText': QtCore.QT_TRANSLATE_NOOP("PartDesign_Sprocket","Sprocket..."),
                'Accel': "",
                'ToolTip': QtCore.QT_TRANSLATE_NOOP("PartDesign_Sprocket","Creates or edit the sprocket definition.")}
        
    def Activated(self):

        FreeCAD.ActiveDocument.openTransaction("Create Sprocket")
        FreeCADGui.addModule("SprocketFeature")
        FreeCADGui.doCommand("SprocketFeature.makeSprocket('Sprocket')")
        FreeCADGui.doCommand("Gui.activeDocument().setEdit(App.ActiveDocument.ActiveObject.Name,0)")
        
    def IsActive(self):
        if FreeCAD.ActiveDocument:
            return True
        else:
            return False

       
class Sprocket:
    """
    The Sprocket object
    """

    def __init__(self,obj):
        self.Type = "Sprocket"
        obj.addProperty("App::PropertyInteger","NumberOfTeeth","Sprocket","Number of gear teeth")
        obj.addProperty("App::PropertyLength","Pitch","Sprocket","Chain Pitch")
        obj.addProperty("App::PropertyLength","RollerDiameter","Sprocket","Roller Diameter")
        obj.addProperty("App::PropertyString","ANSISize","Sprocket","ANSI Size")

        obj.NumberOfTeeth = 50
        obj.Pitch = "0.375 in" 
        obj.RollerDiameter = "0.20 in"
        obj.ANSISize = "35"

        obj.Proxy = self
        
        
    def execute(self,obj):
        w = fcsprocket.FCWireBuilder()
        sprocket.CreateSprocket(w, obj.Pitch.Value, obj.NumberOfTeeth, obj.RollerDiameter.Value)

        sprocketw = Part.Wire([o.toShape() for o in w.wire])
        obj.Shape = sprocketw
        obj.positionBySupport();
        return
        
        
class ViewProviderSprocket:
    """
    A View Provider for the Sprocket object
    """

    def __init__(self,vobj):
        vobj.Proxy = self
       
    def getIcon(self):
        return ":/icons/PartDesign_Sprocket.svg"

    def attach(self, vobj):
        self.ViewObject = vobj
        self.Object = vobj.Object

    def setEdit(self,vobj,mode):
        taskd = SprocketTaskPanel(self.Object,mode)
        taskd.obj = vobj.Object
        taskd.update()
        FreeCADGui.Control.showDialog(taskd)
        return True
    
    def unsetEdit(self,vobj,mode):
        FreeCADGui.Control.closeDialog()
        return

    def __getstate__(self):
        return None

    def __setstate__(self,state):
        return None


class SprocketTaskPanel:
    """
    The editmode TaskPanel for Sprocket objects
    """

    def __init__(self,obj,mode):
        self.obj = obj
        
        self.form=FreeCADGui.PySideUic.loadUi(FreeCAD.getHomePath() + "Mod/PartDesign/SprocketFeature.ui")
        self.form.setWindowIcon(QtGui.QIcon(":/icons/PartDesign_Sprocket.svg"))
        
        QtCore.QObject.connect(self.form.Quantity_Pitch, QtCore.SIGNAL("valueChanged(double)"), self.pitchChanged)
        QtCore.QObject.connect(self.form.Quantity_RollerDiameter, QtCore.SIGNAL("valueChanged(double)"), self.rollerDiameterChanged)
        QtCore.QObject.connect(self.form.spinBox_NumberOfTeeth, QtCore.SIGNAL("valueChanged(int)"), self.numTeethChanged)
        QtCore.QObject.connect(self.form.comboBox_ANSISize, QtCore.SIGNAL("currentTextChanged(const QString)"), self.ANSISizeChanged)
        
        self.update()
        
        if mode == 0: # fresh created
            self.obj.Proxy.execute(self.obj)  # calculate once 
            FreeCAD.Gui.SendMsgToActiveView("ViewFit")
        
    def transferTo(self):
        """
        Transfer from the dialog to the object
        """ 
        self.obj.NumberOfTeeth  = self.form.spinBox_NumberOfTeeth.value()
        self.obj.Pitch        = self.form.Quantity_Pitch.text()
        self.obj.RollerDiameter  = self.form.Quantity_RollerDiameter.text()
        self.obj.ANSISize     = self.form.comboBox_ANSISize.currentText()
    
    def transferFrom(self):
        """
        Transfer from the object to the dialog
        """
        self.form.spinBox_NumberOfTeeth.setValue(self.obj.NumberOfTeeth)
        self.form.Quantity_Pitch.setText(self.obj.Pitch.UserString)
        self.form.Quantity_RollerDiameter.setText(self.obj.RollerDiameter.UserString)
        self.form.comboBox_ANSISize.setCurrentText(self.obj.ANSISize)
                                                    
    def pitchChanged(self, value):
        self.obj.Pitch = value
        self.obj.Proxy.execute(self.obj)
        FreeCAD.Gui.SendMsgToActiveView("ViewFit")

    def ANSISizeChanged(self, size):
        """
        ANSI B29.1-2011 standard roller chain sizes in USCS units (inches)
        {size: [Pitch, Roller Diameter]}
        """
        ANSIRollerTable = {"25": [0.250, 0.130],
                           "35": [0.375, 0.200],
                           "41": [0.500, 0.306],
                           "40": [0.500, 0.312],
                           "50": [0.625, 0.400],
                           "60": [0.750, 0.469],
                           "80": [1.000, 0.625],
                           "100":[1.250, 0.750],
                           "120":[1.500, 0.875],
                           "140":[1.750, 1.000],
                           "160":[2.000, 1.125],
                           "180":[2.250, 1.460],
                           "200":[2.500, 1.562],
                           "240":[3.000, 1.875]}

        self.obj.Pitch           = str(ANSIRollerTable[size][0]) + " in"
        self.obj.RollerDiameter  = str(ANSIRollerTable[size][1]) + " in"
        self.form.Quantity_Pitch.setText(self.obj.Pitch.UserString)
        self.form.Quantity_RollerDiameter.setText(self.obj.RollerDiameter.UserString)
            
        self.obj.Proxy.execute(self.obj)
        FreeCAD.Gui.SendMsgToActiveView("ViewFit")
        
    def rollerDiameterChanged(self, value):
        self.obj.RollerDiameter = value
        self.obj.Proxy.execute(self.obj)

    def numTeethChanged(self, value):
        self.obj.NumberOfTeeth = value
        self.obj.Proxy.execute(self.obj)
        FreeCAD.Gui.SendMsgToActiveView("ViewFit")
        
    def getStandardButtons(self):
        return int(QtGui.QDialogButtonBox.Ok) | int(QtGui.QDialogButtonBox.Cancel)| int(QtGui.QDialogButtonBox.Apply)
    
    def clicked(self,button):
        if button == QtGui.QDialogButtonBox.Apply:
            self.transferTo()
            self.obj.Proxy.execute(self.obj) 
        
    def update(self):
        self.transferFrom()
                
    def accept(self):
        self.transferTo()
        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.ActiveDocument.resetEdit()
        
    def reject(self):
        FreeCADGui.ActiveDocument.resetEdit()
        FreeCAD.ActiveDocument.abortTransaction()



if FreeCAD.GuiUp:
    FreeCADGui.addCommand('PartDesign_Sprocket', CommandSprocket())



# ***************************************************************************
# *   Copyright (c) 2019 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the task panel for the Draft SelectPlane tool."""
## @package task_selectplane
# \ingroup DRAFT
# \brief This module provides the task panel code for the SelectPlane tool.

import FreeCADGui as Gui

# As it is right now this code only loads the task panel .ui file.
# All logic on how to use the widgets is located in the GuiCommand class
# itself.
# On the other hand, the newer tools introduced in v0.19 like OrthoArray,
# PolarArray, and CircularArray include the logic and manipulation
# of the widgets in this task panel class.
# In addition, the task panel code launches the actual function
# using the delayed mechanism defined by the `todo.ToDo` class.
# Therefore, at some point this class should be refactored
# to be more similar to OrthoArray and the new tools.


class SelectPlaneTaskPanel:
    """The task panel definition of the Draft_SelectPlane command."""

    def __init__(self):
        self.form = Gui.PySideUic.loadUi(":/ui/TaskSelectPlane.ui")

    def getStandardButtons(self):
        """Execute to set the standard buttons."""
        return 2097152  # int(QtGui.QDialogButtonBox.Close)



# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""This module provides the object code for Draft make_rectangle function.
"""
## @package make_rectangle
# \ingroup DRAFT
# \brief This module provides the code for Draft make_rectangle function.

import FreeCAD as App

from draftutils.gui_utils import format_object
from draftutils.gui_utils import select

from draftutils.utils import type_check

from draftobjects.polygon import Polygon
from draftviewproviders.view_base import ViewProviderDraft



def make_polygon(nfaces, radius=1, inscribed=True, placement=None, face=None, support=None):
    """makePolgon(edges,[radius],[inscribed],[placement],[face])

    Creates a polygon object with the given number of edges and radius.

    Parameters
    ----------
    edges : int
        Number of edges of the polygon.

    radius : 
        Radius of the control circle.

    inscribed : bool
        Defines is the polygon is inscribed or not into the control circle.

    placement : Base.Placement
        If placement is given, it is used. 
    
    face : bool
        If face is True, the resulting shape is displayed as a face, 
        otherwise as a wireframe.
    
    support : 
        TODO: Describe
    """
    if not App.ActiveDocument:
        App.Console.PrintError("No active document. Aborting\n")
        return
    if nfaces < 3: return None
    obj = App.ActiveDocument.addObject("Part::Part2DObjectPython","Polygon")
    Polygon(obj)
    obj.FacesNumber = nfaces
    obj.Radius = radius
    if face != None:
        obj.MakeFace = face
    if inscribed:
        obj.DrawMode = "inscribed"
    else:
        obj.DrawMode = "circumscribed"
    obj.Support = support
    if placement: obj.Placement = placement
    if App.GuiUp:
        ViewProviderDraft(obj.ViewObject)
        format_object(obj)
        select(obj)

    return obj


makePolygon = make_polygon


# -*- coding: utf-8 -*-

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2018 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from __future__ import print_function

import FreeCAD
import Path
import PathScripts.PathLog as PathLog
import PathScripts.PathOp as PathOp
import PathScripts.PathUtils as PathUtils

from PySide import QtCore

__title__ = "Path Probing Operation"
__author__ = "sliptonic (Brad Collette)"
__url__ = "http://www.freecadweb.org"
__doc__ = "Path Probing operation."

if False:
    PathLog.setLevel(PathLog.Level.DEBUG, PathLog.thisModule())
    PathLog.trackModule(PathLog.thisModule())
else:
    PathLog.setLevel(PathLog.Level.INFO, PathLog.thisModule())


# Qt translation handling
def translate(context, text, disambig=None):
    return QtCore.QCoreApplication.translate(context, text, disambig)


class ObjectProbing(PathOp.ObjectOp):
    '''Proxy object for Probing operation.'''

    def opFeatures(self, obj):
        '''opFeatures(obj) ... Probing works on the stock object.'''
        return PathOp.FeatureDepths | PathOp.FeatureHeights | PathOp.FeatureTool

    def initOperation(self, obj):
        obj.addProperty("App::PropertyLength", "Xoffset", "Probe", QtCore.QT_TRANSLATE_NOOP("App::Property", "X offset between tool and probe"))
        obj.addProperty("App::PropertyLength", "Yoffset", "Probe", QtCore.QT_TRANSLATE_NOOP("App::Property", "Y offset between tool and probe"))
        obj.addProperty("App::PropertyInteger", "PointCountX", "Probe", QtCore.QT_TRANSLATE_NOOP("App::Property", "Number of points to probe in X direction"))
        obj.addProperty("App::PropertyInteger", "PointCountY", "Probe", QtCore.QT_TRANSLATE_NOOP("App::Property", "Number of points to probe in Y direction"))
        obj.addProperty("App::PropertyFile", "OutputFileName", "Path", QtCore.QT_TRANSLATE_NOOP("App::Property", "The output location for the probe data to be written"))

    def nextpoint(self, startpoint=0.0, endpoint=0.0, count=3):
        curstep = 0
        dist = (endpoint - startpoint) / (count - 1)
        while curstep <= count-1:
            yield startpoint + (curstep * dist)
            curstep += 1

    def opExecute(self, obj):
        '''opExecute(obj) ... generate probe locations.'''
        PathLog.track()
        self.commandlist.append(Path.Command("(Begin Probing)"))

        stock = PathUtils.findParentJob(obj).Stock
        bb = stock.Shape.BoundBox

        openstring = '(PROBEOPEN {})'.format(obj.OutputFileName)
        self.commandlist.append(Path.Command(openstring))
        self.commandlist.append(Path.Command("G0", {"Z": obj.ClearanceHeight.Value}))

        for y in self.nextpoint(bb.YMin, bb.YMax, obj.PointCountY):
            for x in self.nextpoint(bb.XMin, bb.XMax, obj.PointCountX):
                self.commandlist.append(Path.Command("G0", {"X": x + obj.Xoffset.Value, "Y": y + obj.Yoffset.Value,  "Z": obj.SafeHeight.Value}))
                self.commandlist.append(Path.Command("G38.2", {"Z": obj.FinalDepth.Value, "F": obj.ToolController.VertFeed.Value}))
                self.commandlist.append(Path.Command("G0", {"Z": obj.SafeHeight.Value}))

        self.commandlist.append(Path.Command("(PROBECLOSE)"))

    def opSetDefaultValues(self, obj, job):
        '''opSetDefaultValues(obj, job) ... set default value for RetractHeight'''


def SetupProperties():
    setup = ['Xoffset', 'Yoffset', 'PointCountX', 'PointCountY', 'OutputFileName']
    return setup


def Create(name, obj=None):
    '''Create(name) ... Creates and returns a Probing operation.'''
    if obj is None:
        obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    proxy = ObjectProbing(obj, name)
    return obj



"""Classes and functions used to test the workbench.

These classes are called by the unit test launcher
that is defined in `Init.py` and `InitGui.py`.

The unit tests are based on the standard `unittest` module.
"""



#***************************************************************************
#*                                                                         *
#*   Copyright (c) 2012 Sebastian Hoogen <github@sebastianhoogen.de>       *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************

__title__="FreeCAD OpenSCAD Workbench - 2D helper functions"
__author__ = "Sebastian Hoogen"
__url__ = ["https://www.freecadweb.org"]

'''
This Script includes python functions to find out the most basic shape type
in a compound and to change the color of shapes according to their shape type
'''

import FreeCAD
def shapedict(shapelst):
    return dict([(shape.hashCode(),shape) for shape in shapelst])

def shapeset(shapelst):
    return set([shape.hashCode() for shape in shapelst])

def mostbasiccompound(comp):
    '''searches for the most basic shape in a Compound'''
    solids=shapeset(comp.Solids)
    shells=shapeset(comp.Shells)
    faces=shapeset(comp.Faces)
    wires=shapeset(comp.Wires)
    edges=shapeset(comp.Edges)
    vertexes=shapeset(comp.Vertexes)
    #FreeCAD.Console.PrintMessage('%s\n' % (str((len(solids),len(shells),len(faces),len(wires),len(edges),len(vertexes)))))
    for shape in comp.Solids:
        shells -= shapeset(shape.Shells)
        faces -= shapeset(shape.Faces)
        wires -= shapeset(shape.Wires)
        edges -= shapeset(shape.Edges)
        vertexes -= shapeset(shape.Vertexes)
    for shape in comp.Shells:
        faces -= shapeset(shape.Faces)
        wires -= shapeset(shape.Wires)
        edges -= shapeset(shape.Edges)
        vertexes -= shapeset(shape.Vertexes)
    for shape in comp.Faces:
        wires -= shapeset(shape.Wires)
        edges -= shapeset(shape.Edges)
        vertexes -= shapeset(shape.Vertexes)
    for shape in comp.Wires:
        edges -= shapeset(shape.Edges)
        vertexes -= shapeset(shape.Vertexes)
    for shape in comp.Edges:
        vertexes -= shapeset(shape.Vertexes)
    #FreeCAD.Console.PrintMessage('%s\n' % (str((len(solids),len(shells),len(faces),len(wires),len(edges),len(vertexes)))))
    #return len(solids),len(shells),len(faces),len(wires),len(edges),len(vertexes)
    if vertexes:
        return "Vertex"
    elif edges:
        return "Edge"
    elif wires:
        return "Wire"
    elif faces:
        return "Face"
    elif shells:
        return "Shell"
    elif solids:
        return "Solid"

def colorcodeshapes(objs):
    shapecolors={
        "Compound":(0.3,0.3,0.4),
        "CompSolid":(0.1,0.5,0.0),
        "Solid":(0.0,0.8,0.0),
        "Shell":(0.8,0.0,0.0),
        "Face":(0.6,0.6,0.0),
        "Wire":(0.1,0.1,0.1),
        "Edge":(1.0,1.0,1.0),
        "Vertex":(8.0,8.0,8.0),
        "Shape":(0.0,0.0,1.0),
        None:(0.0,0.0,0.0)}

    for obj in objs:
        if hasattr(obj,'Shape'):
            try:
                if obj.Shape.isNull():
                    continue
                if not obj.Shape.isValid():
                        color=(1.0,0.4,0.4)
                else:
                    st=obj.Shape.ShapeType
                    if st in ["Compound","CompSolid"]:
                        st = mostbasiccompound(obj.Shape)
                    color=shapecolors[st]
                obj.ViewObject.ShapeColor = color
            except:
                raise

#colorcodeshapes(App.ActiveDocument.Objects)



# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver object Elmer"
__author__ = "Markus Hovorka"
__url__ = "http://www.freecadweb.org"

## \addtogroup FEM
#  @{

from . import tasks
from .equations import elasticity
from .equations import electrostatic
from .equations import flow
from .equations import fluxsolver
from .equations import heat
from .. import run
from .. import solverbase
from femtools import femutils


def create(doc, name="ElmerSolver"):
    return femutils.createObject(
        doc, name, Proxy, ViewProxy)


class Proxy(solverbase.Proxy):
    """Proxy for FemSolverElmers Document Object."""

    Type = "Fem::FemSolverObjectElmer"

    _EQUATIONS = {
        "Heat": heat,
        "Elasticity": elasticity,
        "Electrostatic": electrostatic,
        "Fluxsolver": fluxsolver,
        "Flow": flow,
    }

    def __init__(self, obj):
        super(Proxy, self).__init__(obj)
        obj.addProperty(
            "App::PropertyInteger", "SteadyStateMaxIterations",
            "Steady State", "")
        obj.addProperty(
            "App::PropertyInteger", "SteadyStateMinIterations",
            "Steady State", "")
        obj.addProperty(
            "App::PropertyLink", "ElmerResult",
            "Base", "", 4 | 8)
        obj.addProperty(
            "App::PropertyLink", "ElmerOutput",
            "Base", "", 4 | 8)

        obj.SteadyStateMaxIterations = 1
        obj.SteadyStateMinIterations = 0

    def createMachine(self, obj, directory, testmode=False):
        return run.Machine(
            solver=obj, directory=directory,
            check=tasks.Check(),
            prepare=tasks.Prepare(),
            solve=tasks.Solve(),
            results=tasks.Results(),
            testmode=testmode)

    def createEquation(self, doc, eqId):
        return self._EQUATIONS[eqId].create(doc)

    def isSupported(self, eqId):
        return eqId in self._EQUATIONS


class ViewProxy(solverbase.ViewProxy):
    """Proxy for FemSolverElmers View Provider."""

    def getIcon(self):
        return ":/icons/FEM_SolverElmer.svg"

##  @}



# ***************************************************************************
# *   Copyright (c) 2018 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Tools for FEM unit tests"
__author__ = "Bernd Hahnebach"
__url__ = "http://www.freecadweb.org"

import os
import tempfile
import unittest

import FreeCAD

from os.path import join


def get_fem_test_home_dir(
):
    return join(FreeCAD.getHomePath(), "Mod", "Fem", "femtest", "data")


def get_fem_test_tmp_dir(
):
    temp_dir = join(tempfile.gettempdir(), "FEM_unittests")
    if not os.path.exists(temp_dir):
        os.makedirs(temp_dir)
    return(temp_dir)


def get_unit_test_tmp_dir(
    temp_dir,
    unittestdir
):
    testdir = join(temp_dir, unittestdir)
    if not os.path.exists(testdir):
        os.makedirs(testdir)
    return testdir


def fcc_print(
    message
):
    FreeCAD.Console.PrintMessage("{} \n".format(message))


def get_defmake_count(
    fem_vtk_post=True
):
    """
    count the def make in module ObjectsFem
    could also be done in bash with
    grep -c  "def make" src/Mod/Fem/ObjectsFem.py
    """
    name_modfile = join(FreeCAD.getHomePath(), "Mod", "Fem", "ObjectsFem.py")
    modfile = open(name_modfile, "r")
    lines_modefile = modfile.readlines()
    modfile.close()
    lines_defmake = [l for l in lines_modefile if l.startswith("def make")]
    if not fem_vtk_post:
        # FEM VTK post processing is disabled
        # we are not able to create VTK post objects
        new_lines = []
        for l in lines_defmake:
            if "PostVtk" not in l:
                new_lines.append(l)
        lines_defmake = new_lines
    return len(lines_defmake)


def get_fem_test_defs(
    inout="out"
):
    test_path = join(FreeCAD.getHomePath(), "Mod", "Fem", "femtest")
    collected_test_modules = []
    collected_test_methods = []
    for tfile in sorted(os.listdir(test_path)):
        if tfile.startswith("test") and tfile.endswith(".py"):
            collected_test_modules.append(join(test_path, tfile))
    for f in collected_test_modules:
        tfile = open(f, "r")
        module_name = os.path.splitext(os.path.basename(f))[0]
        class_name = ""
        for ln in tfile:
            ln = ln.lstrip()
            ln = ln.rstrip()
            if ln.startswith("class "):
                ln = ln.lstrip("class ")
                ln = ln.split("(")[0]
                class_name = ln
            if ln.startswith("def test"):
                ln = ln.lstrip("def ")
                ln = ln.split("(")[0]
                collected_test_methods.append(
                    "femtest.{}.{}.{}".format(module_name, class_name, ln)
                )
        tfile.close()
    print("")
    for m in collected_test_methods:
        run_outside_fc = './bin/FreeCADCmd --run-test "{}"'.format(m)
        run_inside_fc = (
            "unittest.TextTestRunner().run(unittest.TestLoader().loadTestsFromName('{}'))"
            .format(m)
        )
        if inout == "in":
            print("\nimport unittest")
            print(run_inside_fc)
        else:
            print(run_outside_fc)


def compare_inp_files(
    file_name1,
    file_name2
):
    file1 = open(file_name1, "r")
    f1 = file1.readlines()
    file1.close()
    # l.startswith("17671.0,1") is a temporary workaround
    # for python3 problem with 1DFlow input
    # TODO as soon as the 1DFlow result reading is fixed
    # this should be triggered in the 1DFlow unit test
    lf1 = [l for l in f1 if not (
        l.startswith("**   written ") or l.startswith("**   file ") or l.startswith("17671.0,1")
    )]
    lf1 = force_unix_line_ends(lf1)
    file2 = open(file_name2, "r")
    f2 = file2.readlines()
    file2.close()
    # TODO see comment on file1
    lf2 = [l for l in f2 if not (
        l.startswith("**   written ") or l.startswith("**   file ") or l.startswith("17671.0,1")
    )]
    lf2 = force_unix_line_ends(lf2)
    import difflib
    diff = difflib.unified_diff(lf1, lf2, n=0)
    result = ""
    for l in diff:
        result += l
    if result:
        result = (
            "Comparing {} to {} failed!\n"
            .format(file_name1, file_name2) + result
        )
    return result


def compare_files(
    file_name1,
    file_name2
):
    file1 = open(file_name1, "r")
    f1 = file1.readlines()
    file1.close()
    # workaround to compare geos of elmer test and temporary file path
    # (not only names change, path changes with operating system)
    lf1 = [l for l in f1 if not (
        l.startswith('Merge "') or l.startswith('Save "') or l.startswith("// ")
    )]
    lf1 = force_unix_line_ends(lf1)
    file2 = open(file_name2, "r")
    f2 = file2.readlines()
    file2.close()
    lf2 = [l for l in f2 if not (
        l.startswith('Merge "') or l.startswith('Save "') or l.startswith("// ")
    )]
    lf2 = force_unix_line_ends(lf2)
    import difflib
    diff = difflib.unified_diff(lf1, lf2, n=0)
    result = ""
    for l in diff:
        result += l
    if result:
        result = "Comparing {} to {} failed!\n".format(file_name1, file_name2) + result
    return result


def compare_stats(
    fea,
    stat_file,
    res_obj_name,
    loc_stat_types=None
):
    import femresult.resulttools as resulttools

    # get the stat types which should be compared
    stat_types = [
        "U1",
        "U2",
        "U3",
        "Uabs",
        "Sabs",
        "MaxPrin",
        "MidPrin",
        "MinPrin",
        "MaxShear",
        "Peeq",
        "Temp",
        "MFlow",
        "NPress"
    ]
    if not loc_stat_types:
        loc_stat_types = stat_types

    # get stats from result obj which should be compared
    obj = fea.analysis.Document.getObject(res_obj_name)
    # fcc_print(obj)
    if obj:
        # fcc_print(obj.Name)
        stats = []
        for s in loc_stat_types:
            statval = resulttools.get_stats(obj, s)
            stats.append(
                "{}: ({:.10f}, {:.10f})\n"
                .format(s, statval[0], statval[1])
            )
    else:
        fcc_print("Result object not found. Name: {}".format(res_obj_name))
        return True

    # get stats to compare with, the expected ones
    sf = open(stat_file, "r")
    sf_content = []
    for l in sf.readlines():
        for st in loc_stat_types:
            if l.startswith(st):
                sf_content.append(l)
    sf.close()
    sf_content = force_unix_line_ends(sf_content)
    if sf_content == []:
        return True

    # compare stats
    if stats != sf_content:
        fcc_print("Stats read from {}.frd file".format(fea.base_name))
        fcc_print("!=")
        fcc_print("Expected stats from {}".format(stat_file))
        for i in range(len(stats)):
            if stats[i] != sf_content[i]:
                fcc_print("{} != {}".format(stats[i].rstrip(), sf_content[i].rstrip()))
        return True

    return False


def force_unix_line_ends(
    line_list
):
    new_line_list = []
    for ln in line_list:
        if ln.endswith("\r\n"):
            ln = ln[:-2] + "\n"
        new_line_list.append(ln)
    return new_line_list


def collect_python_modules(
    femsubdir=None
):
    if not femsubdir:
        pydir = join(FreeCAD.ConfigGet("AppHomePath"), "Mod", "Fem")
    else:
        pydir = join(FreeCAD.ConfigGet("AppHomePath"), "Mod", "Fem", femsubdir)
    collected_modules = []
    fcc_print(pydir)
    for pyfile in sorted(os.listdir(pydir)):
        if pyfile.endswith(".py") and not pyfile.startswith("Init"):
            if not femsubdir:
                collected_modules.append(
                    os.path.splitext(os.path.basename(pyfile))[0]
                )
            else:
                collected_modules.append(
                    femsubdir.replace("/", ".") + "." + os.path.splitext(
                        os.path.basename(pyfile)
                    )[0]
                )
    return collected_modules


def all_test_files(
):
    # open all files
    cube_frequency()
    cube_static()
    Flow1D_thermomech()
    multimat()
    spine_thermomech()


# run the specific test case of the file
# open the file in FreeCAD GUI and return the doc identifier
def cube_frequency(
):
    testname = "femtest.testccxtools.TestCcxTools.test_3_freq_analysis"
    unittest.TextTestRunner().run(unittest.TestLoader().loadTestsFromName(testname))
    doc = FreeCAD.open(join(
        get_fem_test_tmp_dir(),
        "FEM_ccx_frequency",
        "cube_frequency.FCStd")
    )
    return doc


def cube_static(
):
    testname = "femtest.testccxtools.TestCcxTools.test_1_static_analysis"
    unittest.TextTestRunner().run(unittest.TestLoader().loadTestsFromName(testname))
    doc = FreeCAD.open(
        join(get_fem_test_tmp_dir(),
             "FEM_ccx_static",
             "cube_static.FCStd")
    )
    return doc


def Flow1D_thermomech(
):
    testname = "femtest.testccxtools.TestCcxTools.test_5_Flow1D_thermomech_analysis"
    unittest.TextTestRunner().run(unittest.TestLoader().loadTestsFromName(testname))
    doc = FreeCAD.open(join(
        get_fem_test_tmp_dir(),
        "FEM_ccx_Flow1D_thermomech",
        "Flow1D_thermomech.FCStd")
    )
    return doc


def multimat(
):
    testname = "femtest.testccxtools.TestCcxTools.test_2_static_multiple_material"
    unittest.TextTestRunner().run(unittest.TestLoader().loadTestsFromName(testname))
    doc = FreeCAD.open(join(
        get_fem_test_tmp_dir(),
        "FEM_ccx_multimat",
        "multimat.FCStd")
    )
    return doc


def spine_thermomech(
):
    testname = "femtest.testccxtools.TestCcxTools.test_4_thermomech_analysis"
    unittest.TextTestRunner().run(unittest.TestLoader().loadTestsFromName(testname))
    doc = FreeCAD.open(join(
        get_fem_test_tmp_dir(),
        "FEM_ccx_thermomech",
        "spine_thermomech.FCStd")
    )
    return doc



#/***************************************************************************
# *   Copyright (c) Victor Titov (DeepSOIC)                                 *
# *                                           (vv.titov@gmail.com) 2019     *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************/

from Show.SceneDetail import SceneDetail
import FreeCAD as App
import FreeCADGui as Gui

class ClipPlane(SceneDetail):
    """ClipPlane(document, enable = None, placement = None, offset = 0.0):
    TempoVis plugin for applying clipping plane to whole project except edit root.
    enable can be 0 for disable, 1 for enable, and -1 for toggle 
    (FIXME: toggle value support is a hack for while we can't read out the current state)."""
    
    class_id = 'SDClipPlane'
    key = ''
    
    def __init__(self, document, enable = None, placement = None, offset = 0.0):
        self.doc = document
        if enable is not None:
            if placement is not None and offset != 0.0:
                placement = placement.copy()
                dir = placement.multVec(App.Vector(0,0,1))
                placement.Base = placement.Base + dir*offset
            self.data = (enable, placement)
        
    def scene_value(self):
        return (0, None) #hack. For until there is a way to easily query the plane, this should be good enough.
    
    def apply_data(self, val):
        enable, pla = val
        if enable != 0:
            self._viewer().toggleClippingPlane(enable, pla= pla)
        else:
            self._viewer().toggleClippingPlane(enable)

    def _viewer(self):
        if self.doc is None:
            gdoc = Gui.editDocument()
        else:
            gdoc = Gui.getDocument(self.doc.Name)
        v = gdoc.activeView()
        if not hasattr(v, 'toggleClippingPlane'):
            v = gdoc.mdiViewsOfType('Gui::View3DInventor')[0]
        return v



# ***************************************************************************
# *   (c) 2009 Yorik van Havre <yorik@uncreated.net>                        *
# *   (c) 2010 Ken Cline <cline@frii.com>                                   *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides tools for creating Bezier curves with the Draft Workbench.

In particular, a cubic Bezier curve is defined, as it is one of the most
useful curves for many applications.

See https://en.wikipedia.org/wiki/B%C3%A9zier_curve
"""
## @package gui_beziers
# \ingroup DRAFT
# \brief Provides tools for creating Bezier curves with the Draft Workbench.

from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCADGui as Gui
import draftutils.utils as utils
import draftutils.todo as todo
import draftguitools.gui_base_original as gui_base_original
import draftguitools.gui_tool_utils as gui_tool_utils
import draftguitools.gui_lines as gui_lines
import draftguitools.gui_trackers as trackers
from draftutils.messages import _msg, _err
from draftutils.translate import translate


class BezCurve(gui_lines.Line):
    """Gui command for the Bezier Curve tool."""

    def __init__(self):
        super(BezCurve, self).__init__(wiremode=True)
        self.degree = None

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _menu = "Bezier curve"
        _tip = ("Creates an N-degree Bezier curve. "
                "The more points you pick, the higher the degree.\n"
                "CTRL to snap, SHIFT to constrain.")

        return {'Pixmap': 'Draft_BezCurve',
                'Accel': "B, Z",
                'MenuText': QT_TRANSLATE_NOOP("Draft_BezCurve", _menu),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_BezCurve", _tip)}

    def Activated(self):
        """Execute when the command is called.

        Activate the specific bezier curve tracker.
        """
        super(BezCurve, self).Activated(name=translate("draft", "BezCurve"))
        if self.doc:
            self.bezcurvetrack = trackers.bezcurveTracker()

    def action(self, arg):
        """Handle the 3D scene events.

        This is installed as an EventCallback in the Inventor view
        by the `Activated` method of the parent class.

        Parameters
        ----------
        arg: dict
            Dictionary with strings that indicates the type of event received
            from the 3D view.
        """
        if arg["Type"] == "SoKeyboardEvent":
            if arg["Key"] == "ESCAPE":
                self.finish()
        elif arg["Type"] == "SoLocation2Event":  # mouse movement detection
            (self.point,
             ctrlPoint, info) = gui_tool_utils.getPoint(self, arg,
                                                        noTracker=True)

            # existing points + this pointer position
            self.bezcurvetrack.update(self.node + [self.point],
                                      degree=self.degree)
            gui_tool_utils.redraw3DView()
        elif (arg["Type"] == "SoMouseButtonEvent"
              and arg["State"] == "DOWN"
              and arg["Button"] == "BUTTON1"):  # left click
            if arg["Position"] == self.pos:
                self.finish(False, cont=True)

            if (not self.node) and (not self.support):  # first point
                gui_tool_utils.getSupport(arg)
                (self.point,
                 ctrlPoint, info) = gui_tool_utils.getPoint(self, arg,
                                                            noTracker=True)
            if self.point:
                self.ui.redraw()
                self.pos = arg["Position"]
                self.node.append(self.point)  # add point to "clicked list"
                # sb add a control point,
                # if mod(len(cpoints), 2) == 0
                # then create 2 handle points?
                self.drawUpdate(self.point)
                if not self.isWire and len(self.node) == 2:
                    self.finish(False, cont=True)
                if len(self.node) > 2:
                    # does this make sense for a BCurve?
                    # DNC: allows to close the curve
                    # by placing ends close to each other
                    # with tol = Draft tolerance
                    # old code has been to insensitive
                    if (self.point-self.node[0]).Length < utils.tolerance():
                        self.undolast()
                        self.finish(True, cont=True)
                        _msg(translate("draft",
                                       "Bezier curve has been closed"))

    def undolast(self):
        """Undo last line segment."""
        if len(self.node) > 1:
            self.node.pop()
            self.bezcurvetrack.update(self.node, degree=self.degree)
            self.obj.Shape = self.updateShape(self.node)
            _msg(translate("draft", "Last point has been removed"))

    def drawUpdate(self, point):
        """Draw and update to the curve."""
        if len(self.node) == 1:
            self.bezcurvetrack.on()
            if self.planetrack:
                self.planetrack.set(self.node[0])
            _msg(translate("draft", "Pick next point"))
        else:
            self.obj.Shape = self.updateShape(self.node)
            _msg(translate("draft",
                           "Pick next point, "
                           "or finish (A) or close (O)"))

    def updateShape(self, pts):
        """Create shape for display during creation process."""
        import Part
        edges = []
        if len(pts) >= 2:  # allow lower degree segment
            poles = pts[1:]
        else:
            poles = []
        if self.degree:
            segpoleslst = [poles[x:x+self.degree] for x in range(0, len(poles), (self.degree or 1))]
        else:
            segpoleslst = [pts]
        startpoint = pts[0]
        for segpoles in segpoleslst:
            c = Part.BezierCurve()  # last segment may have lower degree
            c.increase(len(segpoles))
            c.setPoles([startpoint] + segpoles)
            edges.append(Part.Edge(c))
            startpoint = segpoles[-1]
        w = Part.Wire(edges)
        return w

    def finish(self, closed=False, cont=False):
        """Terminate the operation and close the curve if asked.

        Parameters
        ----------
        closed: bool, optional
            Close the line if `True`.
        """
        if self.ui:
            if hasattr(self, "bezcurvetrack"):
                self.bezcurvetrack.finalize()
        if not utils.getParam("UiMode", 1):
            Gui.Control.closeDialog()
        if self.obj:
            # remove temporary object, if any
            old = self.obj.Name
            todo.ToDo.delay(self.doc.removeObject, old)
        if len(self.node) > 1:
            # The command to run is built as a series of text strings
            # to be committed through the `draftutils.todo.ToDo` class.
            try:
                rot, sup, pts, fil = self.getStrings()
                Gui.addModule("Draft")
                _cmd = 'Draft.makeBezCurve'
                _cmd += '('
                _cmd += 'points, '
                _cmd += 'closed=' + str(closed) + ', '
                _cmd += 'support=' + sup + ', '
                _cmd += 'degree=' + str(self.degree)
                _cmd += ')'
                _cmd_list = ['points = ' + pts,
                             'bez = ' + _cmd,
                             'Draft.autogroup(bez)',
                             'FreeCAD.ActiveDocument.recompute()']
                self.commit(translate("draft", "Create BezCurve"),
                            _cmd_list)
            except Exception:
                _err("Draft: error delaying commit")

        # `Creator` is the grandfather class, the parent of `Line`;
        # we need to call it to perform final cleanup tasks.
        #
        # Calling it directly like this is a bit messy; maybe we need
        # another method that performs cleanup (superfinish)
        # that is not re-implemented by any of the child classes.
        gui_base_original.Creator.finish(self)
        if self.ui and self.ui.continueMode:
            self.Activated()


Gui.addCommand('Draft_BezCurve', BezCurve())


class CubicBezCurve(gui_lines.Line):
    """Gui command for the 3rd degree Bezier Curve tool."""

    def __init__(self):
        super(CubicBezCurve, self).__init__(wiremode=True)
        self.degree = 3

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _menu = "Cubic bezier curve"
        _tip = ("Creates a Bezier curve made of 2nd degree (quadratic) "
                "and 3rd degree (cubic) segments. "
                "Click and drag to define each segment.\n"
                "After the curve is created you can go back to edit "
                "each control point and set the properties of each knot.\n"
                "CTRL to snap, SHIFT to constrain.")

        return {'Pixmap': 'Draft_CubicBezCurve',
                # 'Accel': "B, Z",
                'MenuText': QT_TRANSLATE_NOOP("Draft_CubicBezCurve", _menu),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_CubicBezCurve", _tip)}

    def Activated(self):
        """Execute when the command is called.

        Activate the specific BezCurve tracker.
        """
        super(CubicBezCurve, self).Activated(name=translate("draft", "CubicBezCurve"))
        if self.doc:
            self.bezcurvetrack = trackers.bezcurveTracker()

    def action(self, arg):
        """Handle the 3D scene events.

        This is installed as an EventCallback in the Inventor view
        by the `Activated` method of the parent class.

        Parameters
        ----------
        arg: dict
            Dictionary with strings that indicates the type of event received
            from the 3D view.
        """
        if arg["Type"] == "SoKeyboardEvent":
            if arg["Key"] == "ESCAPE":
                self.finish()
        elif arg["Type"] == "SoLocation2Event":  # mouse movement detection
            (self.point,
             ctrlPoint, info) = gui_tool_utils.getPoint(self, arg,
                                                        noTracker=True)
            if (len(self.node) - 1) % self.degree == 0 and len(self.node) > 2:
                prevctrl = 2 * self.node[-1] - self.point
                # Existing points + this pointer position
                self.bezcurvetrack.update(self.node[0:-2]
                                          + [prevctrl]
                                          + [self.node[-1]]
                                          + [self.point], degree=self.degree)
            else:
                # Existing points + this pointer position
                self.bezcurvetrack.update(self.node
                                          + [self.point], degree=self.degree)
            gui_tool_utils.redraw3DView()
        elif arg["Type"] == "SoMouseButtonEvent":
            # Press and hold the button
            if arg["State"] == "DOWN" and arg["Button"] == "BUTTON1":
                if arg["Position"] == self.pos:
                    if len(self.node) > 2:
                        self.node = self.node[0:-2]
                    else:
                        self.node = []
                    return
                else:
                    if (not self.node) and (not self.support):  # first point
                        gui_tool_utils.getSupport(arg)
                        (self.point,
                         ctrlPoint,
                         info) = gui_tool_utils.getPoint(self, arg,
                                                         noTracker=True)
                    if self.point:
                        self.ui.redraw()
                        self.pos = arg["Position"]
                        # add point to "clicked list"
                        self.node.append(self.point)
                        # sb add a control point,
                        # if mod(len(cpoints), 2) == 0
                        # then create 2 handle points?
                        self.drawUpdate(self.point)
                        if not self.isWire and len(self.node) == 2:
                            self.finish(False, cont=True)
                        # does this make sense for a BCurve?
                        if len(self.node) > 2:
                            # add point to "clicked list"
                            self.node.append(self.point)
                            self.drawUpdate(self.point)
                            # DNC: allows to close the curve
                            # by placing ends close to each other
                            # with tol = Draft tolerance
                            # old code has been to insensitive
                            _diff = (self.point - self.node[0]).Length
                            if (_diff < utils.tolerance()
                                    and len(self.node) >= 4):
                                # self.undolast()
                                self.node = self.node[0:-2]
                                # close the curve with a smooth symmetric knot
                                _sym = 2 * self.node[0] - self.node[1]
                                self.node.append(_sym)
                                self.finish(True, cont=True)
                                _msg(translate("draft",
                                               "Bezier curve has been closed"))

            # Release the held button
            if arg["State"] == "UP" and arg["Button"] == "BUTTON1":
                if arg["Position"] == self.pos:
                    self.node = self.node[0:-2]
                    return
                else:
                    if (not self.node) and (not self.support):  # first point
                        return
                    if self.point:
                        self.ui.redraw()
                        self.pos = arg["Position"]
                        # add point to "clicked list"
                        self.node.append(self.point)
                        # sb add a control point,
                        # if mod(len(cpoints),2) == 0
                        # then create 2 handle points?
                        self.drawUpdate(self.point)
                        if not self.isWire and len(self.node) == 2:
                            self.finish(False, cont=True)
                        # Does this make sense for a BCurve?
                        if len(self.node) > 2:
                            self.node[-3] = 2 * self.node[-2] - self.node[-1]
                            self.drawUpdate(self.point)
                            # DNC: allows to close the curve
                            # by placing ends close to each other
                            # with tol = Draft tolerance
                            # old code has been to insensitive

    def undolast(self):
        """Undo last line segment."""
        if len(self.node) > 1:
            self.node.pop()
            self.bezcurvetrack.update(self.node, degree=self.degree)
            self.obj.Shape = self.updateShape(self.node)
            _msg(translate("draft", "Last point has been removed"))

    def drawUpdate(self, point):
        """Create shape for display during creation process."""
        if len(self.node) == 1:
            self.bezcurvetrack.on()
            if self.planetrack:
                self.planetrack.set(self.node[0])
            _msg(translate("draft", "Click and drag to define next knot"))
        elif (len(self.node) - 1) % self.degree == 1 and len(self.node) > 2:
            # is a knot
            self.obj.Shape = self.updateShape(self.node[:-1])
            _msg(translate("draft",
                           "Click and drag to define next knot, "
                           "or finish (A) or close (O)"))

    def updateShape(self, pts):
        """Create shape for display during creation process."""
        import Part
        # Not quite right. draws 1 big bez. sb segmented
        edges = []

        if len(pts) >= 2:  # allow lower degree segment
            poles = pts[1:]
        else:
            poles = []

        if self.degree:
            segpoleslst = [poles[x:x+self.degree] for x in range(0, len(poles), (self.degree or 1))]
        else:
            segpoleslst = [pts]

        startpoint = pts[0]

        for segpoles in segpoleslst:
            c = Part.BezierCurve()  # last segment may have lower degree
            c.increase(len(segpoles))
            c.setPoles([startpoint] + segpoles)
            edges.append(Part.Edge(c))
            startpoint = segpoles[-1]
        w = Part.Wire(edges)
        return w

    def finish(self, closed=False, cont=False):
        """Terminate the operation and close the curve if asked.

        Parameters
        ----------
        closed: bool, optional
            Close the line if `True`.
        """
        if self.ui:
            if hasattr(self, "bezcurvetrack"):
                self.bezcurvetrack.finalize()
        if not utils.getParam("UiMode", 1):
            Gui.Control.closeDialog()
        if self.obj:
            # remove temporary object, if any
            old = self.obj.Name
            todo.ToDo.delay(self.doc.removeObject, old)
        if closed is False:
            cleannd = (len(self.node) - 1) % self.degree
            if cleannd == 0:
                self.node = self.node[0:-3]
            if cleannd > 0:
                self.node = self.node[0:-cleannd]
        if len(self.node) > 1:
            try:
                # The command to run is built as a series of text strings
                # to be committed through the `draftutils.todo.ToDo` class.
                rot, sup, pts, fil = self.getStrings()
                Gui.addModule("Draft")
                _cmd = 'Draft.makeBezCurve'
                _cmd += '('
                _cmd += 'points, '
                _cmd += 'closed=' + str(closed) + ', '
                _cmd += 'support=' + sup + ', '
                _cmd += 'degree=' + str(self.degree)
                _cmd += ')'
                _cmd_list = ['points = ' + pts,
                             'bez = ' + _cmd,
                             'Draft.autogroup(bez)',
                             'FreeCAD.ActiveDocument.recompute()']
                self.commit(translate("draft", "Create BezCurve"),
                            _cmd_list)
            except Exception:
                _err("Draft: error delaying commit")

        # `Creator` is the grandfather class, the parent of `Line`;
        # we need to call it to perform final cleanup tasks.
        #
        # Calling it directly like this is a bit messy; maybe we need
        # another method that performs cleanup (superfinish)
        # that is not re-implemented by any of the child classes.
        gui_base_original.Creator.finish(self)
        if self.ui and self.ui.continueMode:
            self.Activated()


Gui.addCommand('Draft_CubicBezCurve', CubicBezCurve())


class BezierGroup:
    """Gui Command group for the Bezier curve tools."""

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _menu = "Bezier tools"
        _tip = ("Create various types of Bezier curves.")

        return {'MenuText': QT_TRANSLATE_NOOP("Draft_BezierTools", _menu),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_BezierTools", _tip)}

    def GetCommands(self):
        """Return a tuple of commands in the group."""
        return ('Draft_CubicBezCurve', 'Draft_BezCurve')

    def IsActive(self):
        """Return True when this command should be available.

        It is `True` when there is a document.
        """
        if Gui.ActiveDocument:
            return True
        else:
            return False


Gui.addCommand('Draft_BezierTools', BezierGroup())
